import pygame
import sys
import os
import socket
import threading
import grpc
from orca_gym.protos.mjc_message_pb2_grpc import GrpcServiceStub 
import orca_gym.protos.mjc_message_pb2 as mjc_message_pb2
import asyncio
import numpy as np

class KeyboardInputSourceType:
    PYGAME = "pygame"
    ORCASTUDIO = "orcastudio"


class KeyboardInputSourcePygame:
    def __init__(self):
        # Initialize Pygame
        pygame.init()

        # Initialize the display (even if not used, it's required for capturing events)
        os.environ['SDL_VIDEO_WINDOW_POS'] = "0,0"
        pygame.display.set_mode((400, 400))

        # Define the key mapping
        self.key_map = {
            pygame.K_a: "A", pygame.K_b: "B", pygame.K_c: "C", pygame.K_d: "D",
            pygame.K_w: "W", pygame.K_s: "S", pygame.K_x: "X", pygame.K_y: "Y", pygame.K_z: "Z",
            pygame.K_q: "Q", pygame.K_e: "E", pygame.K_r: "R", pygame.K_f: "F",
            pygame.K_UP: "Up", pygame.K_DOWN: "Down", pygame.K_LEFT: "Left", pygame.K_RIGHT: "Right",
            pygame.K_SPACE: "Space", pygame.K_RETURN: "Enter", pygame.K_ESCAPE: "Esc",
            pygame.K_LSHIFT: "LShift", pygame.K_RSHIFT: "RShift",
            pygame.K_LCTRL: "Ctrl", pygame.K_RCTRL: "Ctrl",
            pygame.K_LALT: "Alt", pygame.K_RALT: "Alt"
        }

    def update_keyboard_state(self, keyboard_state):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                if event.key in self.key_map:
                    keyboard_state[self.key_map[event.key]] = 1
                    # print(f"Key pressed: {self.key_map[event.key]}")
            elif event.type == pygame.KEYUP:
                if event.key in self.key_map:
                    keyboard_state[self.key_map[event.key]] = 0
                    # print(f"Key released: {self.key_map[event.key]}")

class KeyboardInputSourceOrcaStudio:
    def __init__(self,
                 grpc_addr: str,):
        """
        Initialize the OrcaStudio keyboard input source.
        """
        self.grpc_addr = grpc_addr
        self.loop = asyncio.get_event_loop()
        self.initialize_grpc()

        self.keyboard_map = {
            "keyboard_key_alphanumeric_A" : "A",
            "keyboard_key_alphanumeric_B" : "B",
            "keyboard_key_alphanumeric_C" : "C",
            "keyboard_key_alphanumeric_D" : "D",
            "keyboard_key_alphanumeric_E" : "E",
            "keyboard_key_alphanumeric_F" : "F",
            "keyboard_key_alphanumeric_G" : "G",
            "keyboard_key_alphanumeric_H" : "H",
            "keyboard_key_alphanumeric_I" : "I",
            "keyboard_key_alphanumeric_J" : "J",
            "keyboard_key_alphanumeric_K" : "K",
            "keyboard_key_alphanumeric_L" : "L",
            "keyboard_key_alphanumeric_M" : "M",
            "keyboard_key_alphanumeric_N" : "N",
            "keyboard_key_alphanumeric_O" : "O",
            "keyboard_key_alphanumeric_P" : "P",
            "keyboard_key_alphanumeric_Q" : "Q",
            "keyboard_key_alphanumeric_R" : "R",
            "keyboard_key_alphanumeric_S" : "S",
            "keyboard_key_alphanumeric_T" : "T",
            "keyboard_key_alphanumeric_U" : "U",
            "keyboard_key_alphanumeric_V" : "V",
            "keyboard_key_alphanumeric_W" : "W",
            "keyboard_key_alphanumeric_X" : "X",
            "keyboard_key_alphanumeric_Y" : "Y",
            "keyboard_key_alphanumeric_Z" : "Z",
            "keyboard_key_alphanumeric_0" : "0",
            "keyboard_key_alphanumeric_1" : "1",
            "keyboard_key_alphanumeric_2" : "2",
            "keyboard_key_alphanumeric_3" : "3",
            "keyboard_key_alphanumeric_4" : "4",
            "keyboard_key_alphanumeric_5" : "5",
            "keyboard_key_alphanumeric_6" : "6",
            "keyboard_key_alphanumeric_7" : "7",
            "keyboard_key_alphanumeric_8" : "8",
            "keyboard_key_alphanumeric_9" : "9",
            "keyboard_key_edit_backspace" : "Backspace",
            "keyboard_key_edit_capslock" : "CapsLock",
            "keyboard_key_edit_enter" : "Enter",
            "keyboard_key_edit_space" : "Space",
            "keyboard_key_edit_tab" : "Tab",
            "keyboard_key_escape" : "Esc",
            "keyboard_key_function_F01" : "F1",
            "keyboard_key_function_F02" : "F2",
            "keyboard_key_function_F03" : "F3",
            "keyboard_key_function_F04" : "F4",
            "keyboard_key_function_F05" : "F5",
            "keyboard_key_function_F06" : "F6",
            "keyboard_key_function_F07" : "F7",
            "keyboard_key_function_F08" : "F8",
            "keyboard_key_function_F09" : "F9",
            "keyboard_key_function_F10" : "F10",
            "keyboard_key_function_F11" : "F11",
            "keyboard_key_function_F12" : "F12",
            "keyboard_key_modifier_alt_l" : "Alt",
            "keyboard_key_modifier_alt_r" : "Alt",
            "keyboard_key_modifier_ctrl_l" : "Ctrl",
            "keyboard_key_modifier_ctrl_r" : "Ctrl",
            "keyboard_key_modifier_shift_l" : "LShift",
            "keyboard_key_modifier_shift_r" : "RShift",
            "keyboard_key_navigation_arrow_down" : "Down",
            "keyboard_key_navigation_arrow_left" : "Left",
            "keyboard_key_navigation_arrow_right" : "Right",
            "keyboard_key_navigation_arrow_up" : "Up",
            "mouse_button_left" : "MouseLeft",
            "mouse_button_middle" : "MouseMiddle",
            "mouse_button_right" : "MouseRight",
            "mouse_button_x1" : "MouseX1",
            "mouse_button_x2" : "MouseX2",
        }

    def initialize_grpc(self):
        self.channel = grpc.aio.insecure_channel(
            self.grpc_addr,
        )
        self.stub = GrpcServiceStub(self.channel)

    async def _close_grpc(self):
        if self.channel:
            await self.channel.close()

    async def _get_key_pressed_events(self):
        request = mjc_message_pb2.GetKeyPressedEventsRequest()
        response = await self.stub.GetKeyPressedEvents(request)
        if response is None:
            raise RuntimeError("Failed to get key pressed events from OrcaStudio.")
        return response.events

    def get_key_pressed(self) -> list[str]:
        """
        Get the current key pressed state from OrcaStudio.
        """
        return self.loop.run_until_complete(self._get_key_pressed_events())


    def close(self):
        self.loop.run_until_complete(self._close_grpc())

    def update_keyboard_state(self, keyboard_state : dict[str, int]):
        key_pressed_events = self.get_key_pressed()

        # print(f"Key pressed events: {key_pressed_events}")

        [keyboard_state.update({key: 0}) for key in self.keyboard_map.values()]

        for event in key_pressed_events:
            if event in self.keyboard_map:
                key_name = self.keyboard_map[event]
                if key_name in keyboard_state:
                    keyboard_state[key_name] = 1

class KeyboardInput:
    def __init__(self,
                 input_source: str,
                 grpc_address: str = "localhost:50051"):
        
        # Initialize keyboard input source
        if input_source == KeyboardInputSourceType.PYGAME:
            self._source = KeyboardInputSourcePygame()
        elif input_source == KeyboardInputSourceType.ORCASTUDIO:
            self._source = KeyboardInputSourceOrcaStudio(grpc_address)
        else:
            raise ValueError(f"Unknown input source: {input_source}")

        # Initialize keyboard state
        self.keyboard_state = {
            "W": 0, "A": 0, "S": 0, "D": 0,
            "Space": 0, "LShift": 0, "RShift": 0, "Ctrl": 0, "Alt": 0,
            "Esc": 0, "Enter": 0, "Up": 0, "Down": 0,
            "Left": 0, "Right": 0, "Q": 0, "E": 0, "R": 0, "F": 0, "Z": 0, "X": 0, "Y": 0
        }



    def update(self):
        # Update the keyboard state based on the input source
        self._source.update_keyboard_state(self.keyboard_state)

        # Print the current keyboard state for debugging
        # print(self.keyboard_state)

    def get_state(self):
        return self.keyboard_state.copy()

    def close(self):
        pygame.quit()
        print("Keyboard controller closed")


class KeyboardServer:
    def __init__(self, host='localhost', port=55000):
        self.clients = []
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # 设置 SO_REUSEADDR 选项
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind((host, port))
        except OSError as e:
            print(f"Port {port} is already in use!")
            raise e
        self.server_socket.listen(5)

        pygame.init()
        os.environ['SDL_VIDEO_WINDOW_POS'] = "0,0"        
        self.screen = pygame.display.set_mode((100, 100))
        threading.Thread(target=self.accept_clients).start()

    def accept_clients(self):
        while True:
            try:
                client_socket, addr = self.server_socket.accept()
                self.clients.append(client_socket)
                print(f"Client connected: {addr}")
            except Exception as e:
                print(f"Error accepting clients: {e}")
                break

    def broadcast(self, message):
        for client in self.clients[:]:
            try:
                client.sendall(message.encode())
            except:
                self.clients.remove(client)

    def run(self):
        running = True
        try:
            while running:
                for event in pygame.event.get():
                    if event.type == pygame.KEYDOWN:
                        self.broadcast(f'KEYDOWN:{event.key}')
                    elif event.type == pygame.KEYUP:
                        self.broadcast(f'KEYUP:{event.key}')
                    elif event.type == pygame.QUIT:
                        running = False
        except Exception as e:
            print(f"An error occurred: {e}")
        finally:
            self.close()

    def close(self):
        # 关闭所有客户端套接字
        for client in self.clients:
            client.close()
        self.server_socket.close()
        pygame.quit()
        print("Keyboard server closed")



import socket
import threading
import pygame
import os

class KeyboardClient:
    def __init__(self, host='localhost', port=55000):
        # 初始化键盘状态
        self.keyboard_state = {
            "W": 0, "A": 0, "S": 0, "D": 0,
            "Space": 0, "Shift": 0, "Ctrl": 0, "Alt": 0,
            "Escape": 0, "Enter": 0, "Up": 0, "Down": 0,
            "Left": 0, "Right": 0
        }

        # 定义按键映射
        self.key_map = {
            pygame.K_a: "A", pygame.K_b: "B", pygame.K_c: "C", pygame.K_d: "D",
            pygame.K_w: "W", pygame.K_s: "S", pygame.K_x: "X", pygame.K_y: "Y",
            pygame.K_UP: "Up", pygame.K_DOWN: "Down", pygame.K_LEFT: "Left", pygame.K_RIGHT: "Right",
            pygame.K_SPACE: "Space", pygame.K_RETURN: "Enter", pygame.K_ESCAPE: "Escape"
        }

        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.client_socket.connect((host, port))
        except socket.error as e:
            print(f"无法连接到服务器 {host}:{port}")
            self.client_socket.close()
            raise e

        self.running = True
        self.receive_thread = threading.Thread(target=self.receive)
        self.receive_thread.start()

    def receive(self):
        try:
            while self.running:
                data = self.client_socket.recv(1024).decode()
                if data:
                    event_type, key = data.split(':')
                    key_code = int(key)
                    if key_code in self.key_map:
                        key_name = self.key_map[key_code]
                        if event_type == 'KEYDOWN':
                            self.keyboard_state[key_name] = 1
                        elif event_type == 'KEYUP':
                            self.keyboard_state[key_name] = 0
                else:
                    # 服务器关闭了连接
                    print("服务器关闭了连接")
                    self.running = False
        except Exception as e:
            print(f"接收数据时发生错误: {e}")
        finally:
            self.close()

    def update(self):
        pass

    def get_state(self):
        return self.keyboard_state.copy()
        
    def close(self):
        if self.running:
            self.running = False
            try:
                self.client_socket.shutdown(socket.SHUT_RDWR)
            except Exception as e:
                print(f"关闭套接字时发生错误: {e}")
            finally:
                self.client_socket.close()
                print("键盘客户端已关闭")




import asyncio
import json
import threading
import numpy as np
from scipy.spatial.transform import Rotation as R
import copy

def transform_position_to_mujoco(p_unity):
    x_mujoco = p_unity[2]
    y_mujoco = -p_unity[0]
    z_mujoco = p_unity[1]
    return np.array([x_mujoco, y_mujoco, z_mujoco])

def transform_quaternion_to_mujoco(q_unity):
    R_unity = R.from_quat(q_unity).as_matrix()
    T = np.array([
        [-1, 0, 0],
        [0, 0, 1],
        [0, -1, 0]
    ])
    R_mujoco = T @ R_unity @ T.T
    v = R.from_matrix(R_mujoco).as_rotvec()
    return R.from_rotvec(v).as_quat()

class PicoJoystick:
    def __init__(self, port=8001):
        self.mutex = threading.Lock()
        self.running = True
        self.current_transform = None
        self.current_key_state = None
        self.suceess_AB = 0 #AB键长按次数
        self.suceess_XY = 0 #XY键长按次数
        self.reset_pos = False
        self.loop = asyncio.new_event_loop()
        self.clients = set()  # 初始化 self.clients
        self.port = port
        self.thread = threading.Thread(target=self._start_server_thread)
        self.thread.start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.running = False
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()

    def _start_server_thread(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self._start_server())

    async def _start_server(self):
        server = await asyncio.start_server(
            self._handle_client, '0.0.0.0', self.port)
        async with server:
            await server.serve_forever()

    async def _handle_client(self, reader, writer):
        try:
            self.clients.add(writer)  # 添加客户端 writer
            buffer = ""
            while self.running:
                data = await reader.read(1024)
                if not data:
                    break
                buffer += data.decode('utf-8')
                while '\n' in buffer:
                    line, buffer = buffer.split('\n', 1)
                    message = json.loads(line)
                    with self.mutex:
                        self.current_transform = self.extract_all_transform(message)
                        self.current_key_state = self.extract_key_state(message)
        except Exception as e:
            print("Client disconnected:", e)
        finally:
            self.clients.discard(writer)  # 移除客户端 writer
            writer.close()
            await writer.wait_closed()

    def send_force_message(self, l_hand_force, r_hand_force):
        message = json.dumps({"l_hand_force": l_hand_force, "r_hand_force": r_hand_force})
        asyncio.run_coroutine_threadsafe(self._broadcast_message(message), self.loop)

    async def _broadcast_message(self, message):
        clients = list(self.clients)
        for writer in clients:
            try:
                writer.write((message + '\n').encode('utf-8'))
                await writer.drain()
            except Exception as e:
                print("send message error：", e)
                self.clients.discard(writer)
                writer.close()


    def update(self):
        pass

    def is_reset_pos(self):
        return self.reset_pos

    def set_reset_pos(self, reset_pos):
        self.reset_pos = reset_pos

    def get_transform_list(self):
        with self.mutex:
            if self.current_transform is None:
                return None
            else:
                transform = copy.deepcopy(self.current_transform)
                return transform

    def get_key_state(self):
        with self.mutex:
            if self.current_key_state is None:
                return None
            else:
                key_state = copy.deepcopy(self.current_key_state)
                return key_state

    def extract_single_transform(self, body):
        position = body["position"]
        p_x = position["x"]
        p_y = position["y"]
        p_z = position["z"]
        pos = transform_position_to_mujoco(np.array([p_x, p_y, p_z]))

        rotation = body["rotation"]
        r_x = rotation["x"]
        r_y = rotation["y"]
        r_z = rotation["z"]
        r_w = rotation["w"]
        quat = transform_quaternion_to_mujoco([r_x, r_y, r_z, r_w])

        return [np.array([pos[0], pos[1], pos[2]]), np.array([quat[3], quat[0], quat[1], quat[2]])]

    def extract_all_transform(self, message):
        left_hand_transform = self.extract_single_transform(message["leftHand"])
        right_hand_transform = self.extract_single_transform(message["rightHand"])
        motion_trackers_transform = []
        if "motionTrackers" in message:
            for motionTracker in message["motionTrackers"]:
                motion_trackers_transform.append(self.extract_single_transform(motionTracker))
        return [left_hand_transform, right_hand_transform, motion_trackers_transform]

    def extract_key_state(self, message) -> dict:
        left_hand_key_state = {
            "triggerValue": message["leftHand"]["triggerValue"],
            "primaryButtonPressed": message["leftHand"]["primaryButtonPressed"],
            "secondaryButtonPressed": message["leftHand"]["secondaryButtonPressed"],
            "joystickPosition": [
                message["leftHand"]["joystickPosition"]["x"],
                message["leftHand"]["joystickPosition"]["y"]
            ],
            "joystickPressed": message["leftHand"]["joystickPressed"],
            "gripButtonPressed": message["leftHand"]["gripButtonPressed"]
        }

        right_hand_key_state = {
            "triggerValue": message["rightHand"]["triggerValue"],
            "primaryButtonPressed": message["rightHand"]["primaryButtonPressed"],
            "secondaryButtonPressed": message["rightHand"]["secondaryButtonPressed"],
            "joystickPosition": [
                message["rightHand"]["joystickPosition"]["x"],
                message["rightHand"]["joystickPosition"]["y"]
            ],
            "joystickPressed": message["rightHand"]["joystickPressed"],
            "gripButtonPressed": message["rightHand"]["gripButtonPressed"]
        }
        return {"leftHand": left_hand_key_state, "rightHand": right_hand_key_state}

    def get_right_gripButton(self) -> bool:
        key_state = self.get_key_state()
        if key_state is None:
            return False
        right_hand_key_state = key_state["rightHand"]
        if right_hand_key_state["gripButtonPressed"]:
            return True

        return False
    
    def get_left_gripButton(self) -> bool:
        key_state = self.get_key_state()
        if key_state is None:
            return False
        left_hand_key_state = key_state["leftHand"]
        if left_hand_key_state["gripButtonPressed"]:
            return True

        return False

    def get_success_AB(self) -> bool:
        key_state = self.get_key_state()
        if key_state is None:
            return False
        right_hand_key_state = key_state["rightHand"]
        if right_hand_key_state["primaryButtonPressed"] and right_hand_key_state["secondaryButtonPressed"]:
            self.suceess_AB += 1
        else:
            self.suceess_AB = 0
        return self.suceess_AB >= 20

    def get_success_XY(self) -> bool:
        key_state = self.get_key_state()
        if key_state is None:
            return False
        left_hand_key_state = key_state["leftHand"]
        if left_hand_key_state["primaryButtonPressed"] and left_hand_key_state["secondaryButtonPressed"]:
            self.suceess_XY += 1
        else:
            self.suceess_XY = 0
        return self.suceess_XY >= 20

    def get_left_relative_move(self, transform):
        relative_position = transform[0][0]
        relative_rotation = transform[0][1]
        return [relative_position, relative_rotation]

    def get_right_relative_move(self, transform):
        relative_position = transform[1][0]
        relative_rotation = transform[1][1]
        return [relative_position, relative_rotation]

    def get_motion_trackers_relative_move(self, transform):
        motion_trackers_relative_move = []
        for i in range(len(transform[2])):
            relative_position = transform[2][i][0]
            relative_rotation = transform[2][i][1]
            motion_trackers_relative_move.append([relative_position, relative_rotation])
        return motion_trackers_relative_move

    def get_relative_move(self, transform):
        left_relative_move = self.get_left_relative_move(transform)
        right_relative_move = self.get_right_relative_move(transform)
        motion_trackers_relative_move = self.get_motion_trackers_relative_move(transform)
        return [left_relative_move, right_relative_move, motion_trackers_relative_move]

if __name__ == "__main__":
    with PicoJoystick() as pico_joystick:
        try:
            while True:
                # 主线程可以执行其他任务
                pass
        except KeyboardInterrupt:
            pass  # 上下文管理器会自动调用 close()



import numpy as np

from orca_gym.adapters.robosuite.controllers.base_controller import Controller
from orca_gym.adapters.robosuite.utils.buffers import RingBuffer


class JointVelocityController(Controller):
    """
    Controller for controlling the robot arm's joint velocities. This is simply a P controller with desired torques
    (pre gravity compensation) taken to be proportional to the velocity error of the robot joints.

    NOTE: Control input actions assumed to be taken as absolute joint velocities. A given action to this
    controller is assumed to be of the form: (vel_j0, vel_j1, ... , vel_jn-1) for an n-joint robot

    Args:
        sim (MjSim): Simulator instance this controller will pull robot state updates from

        eef_name (str): Name of controlled robot arm's end effector (from robot XML)

        joint_indexes (dict): Each key contains sim reference indexes to relevant robot joint information, namely:

            :`'joints'`: list of indexes to relevant robot joints
            :`'qpos'`: list of indexes to relevant robot joint positions
            :`'qvel'`: list of indexes to relevant robot joint velocities

        actuator_range (2-tuple of array of float): 2-Tuple (low, high) representing the robot joint actuator range

        input_max (float or list of float): Maximum above which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        input_min (float or list of float): Minimum below which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        output_max (float or list of float): Maximum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        output_min (float or list of float): Minimum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        kp (float or list of float): velocity gain for determining desired torques based upon the joint vel errors.
            Can be either be a scalar (same value for all action dims), or a list (specific values for each dim)

        policy_freq (int): Frequency at which actions from the robot policy are fed into this controller

        velocity_limits (2-list of float or 2-list of list of floats): Limits (m/s) below and above which the magnitude
            of a calculated goal joint velocity will be clipped. Can be either be a 2-list (same min/max value for all
            joint dims), or a 2-list of list (specific min/max values for each dim)

        interpolator (Interpolator): Interpolator object to be used for interpolating from the current joint velocities
            to the goal joint velocities during each timestep between inputted actions

        **kwargs: Does nothing; placeholder to "sink" any additional arguments so that instantiating this controller
            via an argument dict that has additional extraneous arguments won't raise an error
    """

    def __init__(
        self,
        sim,
        eef_name,
        joint_indexes,
        actuator_range,
        input_max=1,
        input_min=-1,
        output_max=1,
        output_min=-1,
        kp=0.25,
        policy_freq=20,
        velocity_limits=None,
        interpolator=None,
        **kwargs,  # does nothing; used so no error raised when dict is passed with extra terms used previously
    ):

        super().__init__(
            sim,
            eef_name,
            joint_indexes,
            actuator_range,
        )
        # Control dimension
        self.control_dim = len(joint_indexes["joints"])

        # input and output max and min (allow for either explicit lists or single numbers)
        self.input_max = self.nums2array(input_max, self.joint_dim)
        self.input_min = self.nums2array(input_min, self.joint_dim)
        self.output_max = self.nums2array(output_max, self.joint_dim)
        self.output_min = self.nums2array(output_min, self.joint_dim)

        # gains and corresopnding vars
        self.kp = self.nums2array(kp, self.joint_dim)
        # if kp is a single value, map wrist gains accordingly (scale down x10 for final two joints)

        if type(kp) is float or type(kp) is int:
            # Scale kpp according to how wide the actuator range is for this robot
            low, high = self.actuator_limits
            self.kp = kp * (high - low)
        self.ki = self.kp * 0.005
        self.kd = self.kp * 0.001
        self.last_err = np.zeros(self.joint_dim)
        self.derr_buf = RingBuffer(dim=self.joint_dim, length=5)
        self.summed_err = np.zeros(self.joint_dim)
        self.saturated = False
        self.last_joint_vel = np.zeros(self.joint_dim)

        # limits
        self.velocity_limits = np.array(velocity_limits) if velocity_limits is not None else None

        # control frequency
        self.control_freq = policy_freq

        # interpolator
        self.interpolator = interpolator

        # initialize torques and goal velocity
        self.goal_vel = None  # Goal velocity desired, pre-compensation
        self.current_vel = np.zeros(self.joint_dim)  # Current velocity setpoint, pre-compensation
        self.torques = None  # Torques returned every time run_controller is called

    def set_goal(self, velocities):
        """
        Sets goal based on input @velocities.

        Args:
            velocities (Iterable): Desired joint velocities

        Raises:
            AssertionError: [Invalid action dimension size]
        """
        # Update state
        self.update()

        # Otherwise, check to make sure velocities is size self.joint_dim
        assert (
            len(velocities) == self.joint_dim
        ), "Goal action must be equal to the robot's joint dimension space! Expected {}, got {}".format(
            self.joint_dim, len(velocities)
        )

        self.goal_vel = self.scale_action(velocities)
        if self.velocity_limits is not None:
            self.goal_vel = np.clip(self.goal_vel, self.velocity_limits[0], self.velocity_limits[1])

        if self.interpolator is not None:
            self.interpolator.set_goal(self.goal_vel)

    def run_controller(self):
        """
        Calculates the torques required to reach the desired setpoint

        Returns:
             np.array: Command torques
        """
        # Make sure goal has been set
        if self.goal_vel is None:
            self.set_goal(np.zeros(self.joint_dim))

        # Update state
        self.update()

        # Only linear interpolator is currently supported
        if self.interpolator is not None:
            if self.interpolator.order == 1:
                # Linear case
                self.current_vel = self.interpolator.get_interpolated_goal()
            else:
                # Nonlinear case not currently supported
                pass
        else:
            self.current_vel = np.array(self.goal_vel)

        # Compute necessary error terms for PID velocity controller
        err = self.current_vel - self.joint_vel
        derr = err - self.last_err
        self.last_err = err
        self.derr_buf.push(derr)

        # Only add to I component if we're not saturated (anti-windup)
        if not self.saturated:
            self.summed_err += err

        # Compute command torques via PID velocity controller plus gravity compensation torques
        torques = self.kp * err + self.ki * self.summed_err + self.kd * self.derr_buf.average + self.torque_compensation

        # Clip torques
        self.torques = self.clip_torques(torques)

        # Check if we're saturated
        self.saturated = False if np.sum(np.abs(self.torques - torques)) == 0 else True

        # Always run superclass call for any cleanups at the end
        super().run_controller()

        # Return final torques
        return self.torques

    def reset_goal(self):
        """
        Resets joint velocity goal to be all zeros
        """
        self.goal_vel = np.zeros(self.joint_dim)

        # Reset interpolator if required
        if self.interpolator is not None:
            self.interpolator.set_goal(self.goal_vel)

    @property
    def name(self):
        return "JOINT_VELOCITY"



import numpy as np

class LowPassFilter:
    def __init__(self, alpha, initial_output):
        self.alpha = alpha
        self.output = initial_output

    def apply(self, input):
        self.output = self.alpha * input + (1 - self.alpha) * self.output
        return self.output


"""
Numba utils.
"""
import numba

import orca_gym.adapters.robosuite.macros as macros


def jit_decorator(func):
    if macros.ENABLE_NUMBA:
        return numba.jit(nopython=True, cache=macros.CACHE_NUMBA)(func)
    return func



import grpc
from orca_gym.protos.mjc_message_pb2_grpc import GrpcServiceStub 
import orca_gym.protos.mjc_message_pb2 as mjc_message_pb2
import asyncio
import numpy as np

class Actor:
    """
    A class to represent an actor in the ORCA Gym environment.
    """

    def __init__(self,
                 name: str,
                 asset_path: str,
                 position: np.ndarray,
                 rotation: np.ndarray,
                 scale: float,):
        self.name = name
        self.asset_path = asset_path
        self.position = position
        self.rotation = rotation
        self.scale = float(scale)
        self._check_actor()

    def _check_actor(self):
        if self.name is None or self.name == "":
            raise ValueError("Actor name cannot be None or empty.")
        if self.asset_path is None or self.asset_path == "":
            raise ValueError("Actor spawnable name cannot be None or empty.")
        if self.position is None or len(self.position) != 3:
            raise ValueError("Actor position must be a 3D vector.")
        if self.rotation is None or len(self.rotation) != 4:
            raise ValueError("Actor rotation must be a quaternion (4D vector).")
        if self.scale is None or not isinstance(self.scale, float):
            raise ValueError("Actor scale must be a float.")

class LightInfo:
    """
    A class to represent light information in the ORCA Gym environment.
    """

    def __init__(self,
                 color: np.ndarray,
                 intensity: float,):
        self.color = color
        self.intensity = intensity
        self._check_light_info()

    def _check_light_info(self):
        if self.color is None or len(self.color) != 3:
            raise ValueError("Light color must be a 3D vector.")
        if self.intensity is None or not isinstance(self.intensity, float):
            raise ValueError("Light intensity must be a float.")
        
class CameraSensorInfo:
    """
    A class to represent camera sensor information in the ORCA Gym environment.
    """

    def __init__(self,
                 capture_rgb : bool,
                 capture_depth : bool,
                 save_mp4_file : bool,
                 use_dds : bool,):
        self.capture_rgb = capture_rgb
        self.capture_depth = capture_depth
        self.save_mp4_file = save_mp4_file
        self.use_dds = use_dds
        self._check_camera_sensor_info()

    def _check_camera_sensor_info(self):
        if self.capture_rgb is None or not isinstance(self.capture_rgb, bool):
            raise ValueError("Capture RGB must be a boolean.")
        if self.capture_depth is None or not isinstance(self.capture_depth, bool):
            raise ValueError("Capture depth must be a boolean.")
        if self.save_mp4_file is None or not isinstance(self.save_mp4_file, bool):
            raise ValueError("Save MP4 file must be a boolean.")
        if self.use_dds is None or not isinstance(self.use_dds, bool):
            raise ValueError("Use DDS must be a boolean.")
        
class MaterialInfo:
    """
    A class to represent material information in the ORCA Gym environment.
    """

    def __init__(self,
                 base_color: np.ndarray,):
        self.base_color = base_color
        self._check_material_info()

    def _check_material_info(self):
        if self.base_color is None or len(self.base_color) != 4:
            raise ValueError("Base color must be a 4D vector.")


class OrcaGymScene:
    """
    A class to represent a scene in the ORCA Gym environment.
    """

    def __init__(self,
                 grpc_addr: str,):
        """
        Initialize the ORCA Gym scene.

        Args:
            scene_id (int): The ID of the scene.
            scene_name (str): The name of the scene.
        """
        self.grpc_addr = grpc_addr
        self.loop = asyncio.get_event_loop()
        self.lock = asyncio.Lock()  # 新增一个异步锁
        self.initialize_grpc()

    def initialize_grpc(self):
        self.channel = grpc.aio.insecure_channel(
            self.grpc_addr,
            options=[
                ('grpc.max_receive_message_length', 1024 * 1024 * 1024),
                ('grpc.max_send_message_length', 1024 * 1024 * 1024),
            ]
        )
        self.stub = GrpcServiceStub(self.channel)

    async def _close_grpc(self):
        if self.channel:
            await self.channel.close()

    def close(self):
        self.loop.run_until_complete(self._close_grpc())

    async def _publish_scene(self):
        async with self.lock:  # 加锁保证串行
            request = mjc_message_pb2.PublishSceneRequest()
            response = await self.stub.PublishScene(request)
            if response.status != mjc_message_pb2.PublishSceneResponse.SUCCESS:
                print("Publish scene failed: ", response.error_message)
                raise Exception("Publish scene failed.")

    def publish_scene(self):
        self.loop.run_until_complete(self._publish_scene())

    async def _add_actor(self, actor : Actor):
        async with self.lock:  # 加锁保证串行
            request = mjc_message_pb2.AddActorRequest(
                name = actor.name,
                spawnable_name = actor.asset_path,
                pos = actor.position,
                quat = actor.rotation,
                scale = actor.scale,)
        
        response = await self.stub.AddActor(request)
        if response.status != mjc_message_pb2.AddActorResponse.SUCCESS:
            print("Add actor failed: ", response.error_message)
            raise Exception("Add actor failed.")
        
    def add_actor(self, actor: Actor):
        self.loop.run_until_complete(self._add_actor(actor))

    async def _set_light_info(self, actor_name : str, light_info: LightInfo):
        async with self.lock:  # 加锁保证串行
            request = mjc_message_pb2.SetLightInfoRequest(
                actor_name = actor_name,
                light_color = light_info.color,
                light_intensity = light_info.intensity,)
            
            response = await self.stub.SetLightInfo(request)
            if response.status != mjc_message_pb2.SetLightInfoResponse.SUCCESS:
                print("Set light info failed: ", response.error_message)
                raise Exception("Set light info failed.")
        
    def set_light_info(self, actor_name : str, light_info: LightInfo):
        self.loop.run_until_complete(self._set_light_info(actor_name, light_info))


    async def _set_camera_sensor_info(self, actor_name: str, camera_sensor_info: CameraSensorInfo):
        async with self.lock:
            request = mjc_message_pb2.SetCameraSensorInfoRequest(
                actor_name = actor_name,
                capture_rgb = camera_sensor_info.capture_rgb,
                capture_depth = camera_sensor_info.capture_depth,
                save_mp4_file = camera_sensor_info.save_mp4_file,
                use_dds = camera_sensor_info.use_dds,)
            
            response = await self.stub.SetCameraSensorInfo(request)
            if response.status != mjc_message_pb2.SetCameraSensorInfoResponse.SUCCESS:
                print("Set camera sensor info failed: ", response.error_message)
                raise Exception("Set camera sensor info failed.")
            
    def set_camera_sensor_info(self, actor_name: str, camera_sensor_info: CameraSensorInfo):
        self.loop.run_until_complete(self._set_camera_sensor_info(actor_name, camera_sensor_info))

    async def _make_camera_viewport_active(self, actor_name: str, entity_name: str):
        async with self.lock:
            request = mjc_message_pb2.MakeCameraViewportActiveRequest(
                actor_name = actor_name,
                entity_name = entity_name,)
            
            response = await self.stub.MakeCameraViewportActive(request)
            if response.status != mjc_message_pb2.MakeCameraViewportActiveResponse.SUCCESS:
                print("Make camera viewport activate failed: ", response.error_message)
                raise Exception("Make camera viewport activate failed.")
            
    def make_camera_viewport_active(self, actor_name: str, entity_name: str):
        self.loop.run_until_complete(self._make_camera_viewport_active(actor_name, entity_name))

    async def _set_material_info(self, actor_name: str, material_info: MaterialInfo):
        async with self.lock:
            request = mjc_message_pb2.SetMaterialInfoRequest(
                actor_name = actor_name,
                base_color = material_info.base_color,)
            
            response = await self.stub.SetMaterialInfo(request)
            if response.status != mjc_message_pb2.SetMaterialInfoResponse.SUCCESS:
                print("Set material info failed: ", response.error_message)
                raise Exception("Set material info failed.")
            
    def set_material_info(self, actor_name: str, material_info: MaterialInfo):
        self.loop.run_until_complete(self._set_material_info(actor_name, material_info))


    async def _set_actor_anim_param_number(self, actor_name: str, param_name: str, value: float):
        async with self.lock:
            request = mjc_message_pb2.SetActorAnimParamNumberRequest(
                actor_name = actor_name,
                param_name = param_name,
                value = value,)
            
            response = await self.stub.SetActorAnimParamNumber(request)
            if response.status != mjc_message_pb2.SetActorAnimParamNumberResponse.SUCCESS:
                print("Set actor anim param number failed: ", response.error_message)
                raise Exception("Set actor anim param number failed.")
            
    def set_actor_anim_param_number(self, actor_name: str, param_name: str, value: float):
        self.loop.run_until_complete(self._set_actor_anim_param_number(actor_name, param_name, value))

    async def _set_actor_anim_param_bool(self, actor_name: str, param_name: str, value: bool):
        async with self.lock:
            request = mjc_message_pb2.SetActorAnimParamBoolRequest(
                actor_name = actor_name,
                param_name = param_name,
                value = value,)
            
            response = await self.stub.SetActorAnimParamBool(request)
            if response.status != mjc_message_pb2.SetActorAnimParamBoolResponse.SUCCESS:
                print("Set actor anim param bool failed: ", response.error_message)
                raise Exception("Set actor anim param bool failed.")
            
    def set_actor_anim_param_bool(self, actor_name: str, param_name: str, value: bool):
        self.loop.run_until_complete(self._set_actor_anim_param_bool(actor_name, param_name, value))

    async def _set_actor_anim_param_string(self, actor_name: str, param_name: str, value: str):
        async with self.lock:
            request = mjc_message_pb2.SetActorAnimParamStringRequest(
                actor_name = actor_name,
                param_name = param_name,
                value = value,)
            
            response = await self.stub.SetActorAnimParamString(request)
            if response.status != mjc_message_pb2.SetActorAnimParamStringResponse.SUCCESS:
                print("Set actor anim param string failed: ", response.error_message)
                raise Exception("Set actor anim param string failed.")
            
    def set_actor_anim_param_string(self, actor_name: str, param_name: str, value: str):
        self.loop.run_until_complete(self._set_actor_anim_param_string(actor_name, param_name, value))


import os
import shutil
import sys

def get_absolute_base_path():
    """获取绝对基础路径(OrcaGym目录)"""
    # 获取脚本绝对路径
    script_path = os.path.abspath(__file__)
    # 向上回溯3级目录到OrcaGym
    base_path = os.path.dirname(os.path.dirname(os.path.dirname(script_path)))
    return base_path

def main():
    # 检查参数
    if len(sys.argv) != 3:
        print("用法: python rename_dirs.py <相对路径> <目标目录名>")
        print("示例: python rename_dirs.py examples/openpi/records_tmp/ shop")
        sys.exit(1)
    
    relative_path = sys.argv[1].strip('/')  # 去除可能的末尾斜杠
    target_dir_name = sys.argv[2]
    
    # 计算基础路径
    base_path = get_absolute_base_path()
    print(f"[调试] 计算出的基础路径: {base_path}")
    
    # 构建完整目标路径
    target_path = os.path.join(base_path, relative_path, target_dir_name)
    print(f"[调试] 完整目标路径: {target_path}")
    
    # 验证路径是否存在
    if not os.path.exists(target_path):
        print(f"错误: 目标目录不存在: {target_path}")
        print("可能原因:")
        print(f"1. 相对路径 '{relative_path}' 不正确")
        print(f"2. 目标目录名 '{target_dir_name}' 不正确")
        print("请检查:")
        print(f"   - 脚本所在目录: {os.path.dirname(os.path.abspath(__file__))}")
        print(f"   - 计算出的基础路径: {base_path}")
        print(f"   - 尝试访问的路径: {target_path}")
        sys.exit(1)
    
    # 确认是目录
    if not os.path.isdir(target_path):
        print(f"错误: 目标路径不是目录: {target_path}")
        sys.exit(1)
    
    # 处理目录
    try:
        # 1. 获取子场景目录 (如 pick_catch)
        subscene_dirs = [d for d in os.listdir(target_path) if os.path.isdir(os.path.join(target_path, d))]
        if len(subscene_dirs) != 1:
            print(f"错误: {target_path} 下子场景目录数量不正确(应为1个，实际找到{len(subscene_dirs)}个)")
            print(f"找到的子目录: {subscene_dirs}")
            sys.exit(1)
        
        subscene_dir = os.path.join(target_path, subscene_dirs[0])
        print(f"[调试] 找到子场景目录: {subscene_dir}")
        
        # 2. 获取动作目录
        action_dirs = [d for d in os.listdir(subscene_dir) if os.path.isdir(os.path.join(subscene_dir, d))]
        if not action_dirs:
            print(f"错误: {subscene_dir} 下没有动作目录")
            sys.exit(1)
        print(f"[调试] 找到动作目录: {action_dirs}")
        
        # 3. 解析第一个动作目录名获取大小信息
        first_action = action_dirs[0]
        parts = first_action.split("_")
        if len(parts) < 4:
            print(f"错误: {first_action} 目录名格式不正确")
            print(f"目录名应包含至少3个下划线分隔的部分，如: 名称_53p21GB_2000counts_85p30h")
            sys.exit(1)
        
        size_part, files_part, duration_part = parts[-3], parts[-2], parts[-1]
        print(f"[调试] 解析出的大小信息: {size_part}, {files_part}, {duration_part}")
        
        # 4. 创建新目录名
        new_target_name = f"{target_dir_name}-{size_part}_{files_part}_{duration_part}"
        new_subscene_name = f"{subscene_dirs[0]}-{size_part}_{files_part}_{duration_part}"
        print(f"[调试] 新目录名: {new_target_name}/{new_subscene_name}")
        
        # 5. 创建新目录结构
        parent_dir = os.path.dirname(target_path)
        new_target_dir = os.path.join(parent_dir, new_target_name)
        new_subscene_dir = os.path.join(new_target_dir, new_subscene_name)
        os.makedirs(new_subscene_dir, exist_ok=True)
        print(f"[调试] 创建新目录结构: {new_target_dir}/{new_subscene_dir}")
        
        # 6. 移动并重命名动作目录
        for action in action_dirs:
            new_action_name = f"{action}-{size_part}_{files_part}_{duration_part}"
            src_path = os.path.join(subscene_dir, action)
            dst_path = os.path.join(new_subscene_dir, new_action_name)
            print(f"移动: {src_path} -> {dst_path}")
            shutil.move(src_path, dst_path)
        
        # 7. 删除空目录
        os.rmdir(subscene_dir)
        os.rmdir(target_path)
        print(f"成功: {target_path} 已重命名为 {new_target_dir}/{new_subscene_dir}/...")
        
    except Exception as e:
        print(f"错误: 处理过程中发生异常: {str(e)}")
        sys.exit(1)

if __name__ == "__main__":
    main()


import random
import os
import sys
import argparse

def generate_geom_terrain(num_x, num_y, geom_type, geom_size, geom_size_cale_range, max_tilt, min_step, max_step, max_total_height, min_spacing, max_spacing, rotation_z_min, rotation_z_max):
    terrain = []
    heights = [[0 for _ in range(num_y)] for _ in range(num_x)]
    prev_color = random.uniform(0.1, 0.5)  # 初始化第一个box的颜色，取值范围在0.1到0.5之间

    for i in range(num_x):
        for j in range(num_y):
            if i == 0 and j == 0:
                heights[i][j] = 0  # 第一块高度为 0
            else:
                # 计算相邻高度差，限制在 min_step 到 max_step 之间
                if i > 0 and j > 0:
                    height = min(heights[i-1][j], heights[i][j-1])
                elif i > 0:
                    height = heights[i-1][j]
                else:
                    height = heights[i][j-1]

                if random.uniform(0, 1) < 0.5:
                    delta_h = random.uniform(-min_step * geom_size[2], -max_step * geom_size[2])  # 限制高度差在几何体高度的 min_step 到 max_step 之间
                else:
                    delta_h = random.uniform(min_step * geom_size[2], max_step * geom_size[2])  # 限制高度差在几何体高度的 min_step 到 max_step 之间
                heights[i][j] = max(0, min(max_total_height, height + delta_h))

            # 随机生成倾斜角度和旋转
            tilt_x = random.uniform(-max_tilt, max_tilt)
            tilt_y = random.uniform(-max_tilt, max_tilt)
            rotation_z = random.uniform(rotation_z_min, rotation_z_max)  # 旋转角度范围参数化

            # 生成中心间距的随机值
            spacing_x = random.uniform(min_spacing * geom_size[0], max_spacing * geom_size[0])
            spacing_y = random.uniform(min_spacing * geom_size[0], max_spacing * geom_size[0])
            
            pos_x = i * spacing_x
            pos_y = j * spacing_y
            pos_z = heights[i][j] # + geom_size[2]  # 让box稍微浮出地面一点

            # 随机生成灰度颜色，确保相邻块色差大于 0.2，灰度范围在0.1到0.5之间
            while True:
                gray_value = random.uniform(0.1, 0.5)
                if abs(gray_value - prev_color) > 0.2:
                    break
            
            prev_color = gray_value  # 更新前一个颜色值
            rgba = f"{gray_value} {gray_value} {gray_value} 1"

            # 构建地形的box geom定义
            size_x = (geom_size[0] + geom_size[0] * random.uniform(-geom_size_cale_range, geom_size_cale_range)) / 2
            size_y = (geom_size[1] + geom_size[1] * random.uniform(-geom_size_cale_range, geom_size_cale_range)) / 2
            size_z = (geom_size[2] + geom_size[2] * random.uniform(-geom_size_cale_range, geom_size_cale_range)) / 2
            terrain.append(f'<geom type="{geom_type}" pos="{pos_x} {pos_y} {pos_z}" size="{size_x} {size_y} {size_z}" euler="{tilt_x} {tilt_y} {rotation_z}" rgba="{rgba}"/>')

    return terrain

def generate_mujoco_xml(num_x, num_y, geom_type, geom_size, geom_size_cale_range, max_tilt, min_step, max_step, max_total_height, min_spacing, max_spacing, rotation_z_min, rotation_z_max, out_file):
    # 生成地形的所有geom块
    terrain_boxes = generate_geom_terrain(num_x, num_y, geom_type, geom_size, geom_size_cale_range, max_tilt, min_step, max_step, max_total_height, min_spacing, max_spacing, rotation_z_min, rotation_z_max)
    terrain_center = (num_x * -geom_size[0] / 2, num_y * -geom_size[1] / 2, 0)

    # 定义MuJoCo XML的头部和尾部
    header = f'''<mujoco model="random_box_terrain">
    <compiler angle="degree" coordinate="local"/>
    <option timestep="0.01" gravity="0 0 -9.81" integrator="RK4"/>

    <worldbody>
        <body name="terrain" pos="{terrain_center[0]} {terrain_center[1]} {terrain_center[2]}">
    '''

    footer = '''
        </body>
    </worldbody>

    <actuator>
        <!-- 在这里你可以定义一些控制器 -->
    </actuator>

</mujoco>'''

    # 拼接完整的XML文件
    xml_content = header + "\n".join(terrain_boxes) + footer

    # 将XML内容写入文件
    with open(f"{out_file}", "w") as f:
        f.write(xml_content)

    print(f"XML文件已生成为 {out_file}")

def parse_geom_size(s):
    try:
        # 移除括号并分割字符串
        s = s.strip('()')
        parts = s.split(',')
        if len(parts) != 3:
            raise ValueError
        return tuple(float(part) for part in parts)
    except:
        raise argparse.ArgumentTypeError("geom_size must be a tuple of three floats, e.g., (0.5,0.3,0.2)")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Run multiple instances of the script with different gRPC addresses.')
    parser.add_argument('--num_x', type=int, default=16, help='Number of boxes in x direction')
    parser.add_argument('--num_y', type=int, default=16, help='Number of boxes in y direction')
    parser.add_argument('--geom_type', type=str, default='box', help='Type of geom to use for terrain generation (sphere, ellipsoid, box, cylinder, capsule)')    
    parser.add_argument('--geom_size', type=parse_geom_size, default="(1, 1, 0.2)", help='Size (x, y, z) of each geom')
    parser.add_argument('--geom_size_cale_range', type=float, default=0.5, help="The random range of geom size")
    parser.add_argument('--max_tilt', type=float, default=3, help='Max tilt of each geom')
    parser.add_argument('--min_step', type=float, default=0.2, help='Min step between geoms')
    parser.add_argument('--max_step', type=float, default=0.5, help='Max step between geoms')
    parser.add_argument('--max_total_height', type=float, default=1, help='Max total height of terrain')
    parser.add_argument('--min_spacing', type=float, default=1, help='Min spacing between geoms')
    parser.add_argument('--max_spacing', type=float, default=1.5, help='Max spacing between geoms')
    parser.add_argument('--rotation_z_min', type=float, default=0, help='Min rotation around z axis')
    parser.add_argument('--rotation_z_max', type=float, default=360, help='Max rotation around z axis')
    parser.add_argument('--output', type=str, default='terrain.xml', help='Output file name')
    args = parser.parse_args()


    num_x = args.num_x
    num_y = args.num_y
    geom_type = args.geom_type
    geom_size = args.geom_size
    geom_size_cale_range = args.geom_size_cale_range
    max_tilt = args.max_tilt
    min_step = args.min_step
    max_step = args.max_step
    max_total_height = args.max_total_height
    min_spacing = args.min_spacing
    max_spacing = args.max_spacing
    rotation_z_min = args.rotation_z_min
    rotation_z_max = args.rotation_z_max
    out_file = args.output

    generate_mujoco_xml(num_x, num_y, geom_type, geom_size, geom_size_cale_range, max_tilt, min_step, max_step, 
                        max_total_height, min_spacing, max_spacing, rotation_z_min, rotation_z_max, out_file)

    # generate_mujoco_xml(
    #     10,   # num_x: x方向的box数量
    #     10,   # num_y: y方向的box数量
    #     "ellipsoid",  # geom_type: box的类型
    #     1,    # box_size: 每个box的边长 (2米 x 2米)
    #     3,    # max_tilt: box相对于水平面的最大倾斜角 (3度)
    #     -0.05,  # min_step: 相邻box之间的最小高度差 (0.1米)
    #     0.05,  # max_step: 相邻box之间的最大高度差 (0.2米)
    #     1,    # max_total_height: 场景的总高差最大为1米
    #     0.75,  # min_spacing: box之间的最小中心间距 (0.25米)
    #     1,    # max_spacing: box之间的最大中心间距 (0.5米)
    #     0,    # rotation_z_min: box相对于垂直方向的最小旋转角度
    #     360   # rotation_z_max: box相对于垂直方向的最大旋转角度
    # )



from orca_gym.scene.orca_gym_scene import OrcaGymScene, Actor, LightInfo, CameraSensorInfo, MaterialInfo
from orca_gym.scene.orca_gym_scene_runtime import OrcaGymSceneRuntime
import numpy as np
import orca_gym.utils.rotations as rotations
import time
import random
import gymnasium as gym
import sys
from datetime import datetime
import os
from typing import Optional

ENV_ENTRY_POINT = {
    "SimulationLoop": "orca_gym.scripts.sim_env:SimEnv",
}

TIME_STEP = 0.001
FRAME_SKIP = 20
REALTIME_STEP = TIME_STEP * FRAME_SKIP
CONTROL_FREQ = 1 / REALTIME_STEP

def register_env(orcagym_addr : str, 
                 env_name : str, 
                 env_index : int, 
                 agent_name : str, 
                 max_episode_steps : int) -> tuple[ str, dict ]:
    orcagym_addr_str = orcagym_addr.replace(":", "-")
    env_id = env_name + "-OrcaGym-" + orcagym_addr_str + f"-{env_index:03d}"
    agent_names = [f"{agent_name}"]
    kwargs = {'frame_skip': FRAME_SKIP,   
                'orcagym_addr': orcagym_addr, 
                'agent_names': agent_names, 
                'time_step': TIME_STEP}           
    gym.register(
        id=env_id,
        entry_point=ENV_ENTRY_POINT[env_name],
        kwargs=kwargs,
        max_episode_steps= max_episode_steps,
        reward_threshold=0.0,
    )
    return env_id, kwargs



def run_simulation(orcagym_addr : str, 
                agent_name : str,
                env_name : str,
                scene_runtime: Optional[OrcaGymSceneRuntime] = None) -> None:
    env = None  # Initialize env to None
    try:
        print("simulation running... , orcagym_addr: ", orcagym_addr)

        env_index = 0
        env_id, kwargs = register_env(orcagym_addr, 
                                      env_name, 
                                      env_index, 
                                      agent_name, 
                                      sys.maxsize)
        print("Registered environment: ", env_id)

        env = gym.make(env_id)        
        print("Starting simulation...")

        if scene_runtime is not None:
            if hasattr(env, "set_scene_runtime"):
                print("Setting scene runtime...")
                env.set_scene_runtime(scene_runtime)
            elif hasattr(env.unwrapped, "set_scene_runtime"):
                print("Setting scene runtime...")
                env.unwrapped.set_scene_runtime(scene_runtime)

        obs = env.reset()
        while True:
            start_time = datetime.now()

            action = env.action_space.sample()
    
            obs, reward, terminated, truncated, info = env.step(action)

            env.render()

            elapsed_time = datetime.now() - start_time
            if elapsed_time.total_seconds() < REALTIME_STEP:
                time.sleep(REALTIME_STEP - elapsed_time.total_seconds())


    except KeyboardInterrupt:
        print("Simulation stopped")        
        if env is not None:
            env.close()


if __name__ == "__main__":
    orcagym_addr = "localhost:50051"
    agent_name = "NoRobot"
    env_name = "SimulationLoop"
    run_simulation(orcagym_addr, agent_name, env_name)



import cv2
import robomimic.envs.env_base as EB
import h5py
import numpy as np
import json
import os
import uuid  # 新增导入uuid模块
import time  # 新增导入time模块
import shutil
import hashlib
class DatasetWriter:
    # def __init__(self, base_dir, env_name, env_version, env_kwargs=None):  # 修改1: 参数名改为base_dir
    #     """
    #     初始化 DatasetWriter。

    #     参数：
    #     - base_dir: 基础目录路径，用于创建UUID子目录
    #     - env_name: 环境名称。
    #     - env_version: 环境版本。
    #     - env_kwargs: 环境参数字典（可选）。
    #     """
    #     self.experiment_id = str(uuid.uuid4())[:8]  # 修改2: 使用8位UUID简化路径
    #     self.uuid_dir = os.path.join(base_dir, f"{self.experiment_id}_{int(time.time())}")  # 修改3: 结合时间戳
    #     os.makedirs(self.uuid_dir, exist_ok=True)  # 新增: 创建UUID目录
        
    #     self.mp4_save_path : str = None

    #     self._env_args = {
    #         "env_name": env_name,
    #         "type": EB.EnvType.ORCA_GYM_TYPE,
    #         "env_version": env_version,
    #         "env_kwargs": env_kwargs or {}
    #     }

    #     # 创建新的 HDF5 文件，写入初始数据，然后关闭文件
    #     self.hdf5_path = os.path.join(self.uuid_dir, "env_data.hdf5")  # 新增: 定义HDF5文件路径
    #     with h5py.File(self.hdf5_path, 'w') as f:  # 修改4: 使用hdf5_path
    #         data_group = f.create_group('data')
    #         data_group.attrs['env_args'] = json.dumps(self._env_args)
    #         data_group.attrs['total'] = 0  # 初始化总样本数为 0
    #         data_group.attrs['demo_count'] = 0  # 初始化演示计数为 0
    #         f.create_group('mask')  # 创建掩码组用于过滤器键（可选）
    def __init__(self, base_dir, env_name, env_version, env_kwargs=None, specific_file_path=None):  # 修改1: 参数名改为base_dir
        """
        初始化 DatasetWriter。

        参数：
        - base_dir: 基础目录路径，用于创建UUID子目录
        - env_name: 环境名称。
        - env_version: 环境版本。
        - env_kwargs: 环境参数字典（可选）。
        """

        self._env_args = {
            "env_name": env_name,
            "type": EB.EnvType.ORCA_GYM_TYPE,
            "env_version": env_version,
            "env_kwargs": env_kwargs or {}
        }
        self.pathremoved = False
       # print("base_dir..................:", base_dir)
        if specific_file_path is not None:  # 新增: 如果提供了特定文件路径
            self.hdf5_path = specific_file_path  # 使用特定文件路径
            self.create_hdf5_file()  # 新增: 创建HDF5文件
        else:
            self.basedir = base_dir
            # self.experiment_id = str(uuid.uuid4())[:8]  # 修改2: 使用8位UUID简化路径
            # self.uuid_dir = os.path.join(base_dir, f"{self.experiment_id}_{int(time.time())}")  # 修改3: 结合时间戳

            # uuid1 = str(uuid.uuid4())[:8]
            # uuid2 = str(uuid.uuid4())[:8]
            raw_uuid = str(uuid.uuid4())
            hash_digest = hashlib.sha256(raw_uuid.encode('utf-8')).hexdigest()
            short_id = hash_digest[:16]
            self.experiment_id = short_id
            # self.experiment_id = f"{uuid1}_{uuid2}"
            self.uuid_dir = os.path.join(base_dir, self.experiment_id)
            self.camera_dir = os.path.join(self.uuid_dir, "camera")
            self.parameters_dir = os.path.join(self.uuid_dir, "parameters")
            self.proprio_stats_dir = os.path.join(self.uuid_dir, "proprio_stats")
            self.depth_dir = os.path.join(self.camera_dir, "depth")
            self.video_dir = os.path.join(self.camera_dir, "video")
            print("self.uuid_dir..............:",self.uuid_dir)
            self.hdf5_path = os.path.join(self.proprio_stats_dir, "proprio_stats.hdf5") 
            self.mp4_save_path : str = None
            
        



    def create_hdf5_file(self):
        with h5py.File(self.hdf5_path, 'w') as f:  # 修改4: 使用hdf5_path
            data_group = f.create_group('data')
            data_group.attrs['env_args'] = json.dumps(self._env_args)
            data_group.attrs['total'] = 0  # 初始化总样本数为 0
            data_group.attrs['demo_count'] = 0  # 初始化演示计数为 0
            f.create_group('mask')  # 创建掩码组用于过滤器键（可选）

    def set_UUIDPATH(self):
        # self.experiment_id = str(uuid.uuid4())[:8]  # 修改2: 使用8位UUID简化路径
        # self.uuid_dir = os.path.join(self.basedir, f"{self.experiment_id}_{int(time.time())}")  # 修改3: 结合时间戳
        # uuid1 = str(uuid.uuid4())[:8]
        # uuid2 = str(uuid.uuid4())[:8]
        # self.experiment_id = f"{uuid1}_{uuid2}"
        raw_uuid = str(uuid.uuid4())
        hash_digest = hashlib.sha256(raw_uuid.encode('utf-8')).hexdigest()
        short_id = hash_digest[:16]
        self.experiment_id = short_id
        self.uuid_dir = os.path.join(self.basedir, self.experiment_id)
        self.camera_dir = os.path.join(self.uuid_dir, "camera")
        self.proprio_stats_dir = os.path.join(self.uuid_dir, "proprio_stats")
        self.depth_dir = os.path.join(self.camera_dir, "depth")
        self.video_dir = os.path.join(self.camera_dir, "video")
        self.parameters_dir = os.path.join(self.uuid_dir, "parameters")
       # print("self.uuid_dir:",self.uuid_dir)
        os.makedirs(self.uuid_dir, exist_ok=True)  # 新增: 创建UUID目录
        os.makedirs(self.camera_dir, exist_ok=True)
        os.makedirs(self.parameters_dir, exist_ok=True)
        os.makedirs(self.proprio_stats_dir, exist_ok=True)
        os.makedirs(self.depth_dir, exist_ok=True)
        os.makedirs(self.video_dir, exist_ok=True)
        # 创建新的 HDF5 文件，写入初始数据，然后关闭文件
        self.hdf5_path = os.path.join(self.proprio_stats_dir, "proprio_stats.hdf5")  # 新增: 定义HDF5文件路径
        # self.hdf5_path = os.path.join(self.uuid_dir)
        # print("hdf5_path11112222:",self.hdf5_path)
        self.create_hdf5_file()  # 新增: 创建HDF5文件
        self.pathremoved = False
    
    def get_UUIDPath(self):
        return self.uuid_dir

    def remove_path(self):
        rmpath = self.uuid_dir
        print(f"删除目录: {rmpath}")
        if os.path.exists(rmpath) and os.path.isdir(rmpath):
            shutil.rmtree(rmpath)
        self.pathremoved = True


    @staticmethod
    def remove_episode_from_json(json_path, episode_id):
    # 删除 JSON 文件中指定 episode_id 的记录
        if not os.path.exists(json_path):
            print(f"JSON file {json_path} not found!")
            return

        with open(json_path, "r", encoding="utf-8") as f:
            data = json.load(f)

        # 保留不是这个 episode_id 的条目
        new_data = [item for item in data if item["episode_id"] != episode_id]

        with open(json_path, "w", encoding="utf-8") as f:
            json.dump(new_data, f, ensure_ascii=False, indent=4)
        
        print(f"Removed episode {episode_id} from {json_path}")


        


    def set_env_kwargs(self, env_kwargs):
        """
        设置环境参数。

        参数：
        - env_kwargs: 环境参数字典。
        """
        self._env_args['env_kwargs'] = env_kwargs
        with h5py.File(self.hdf5_path, 'r+') as f:  # 修改5: 使用hdf5_path
            data_group = f['data']
            data_group.attrs['env_args'] = json.dumps(self._env_args)

    def get_cur_demo_name(self) -> str:
        with h5py.File(self.hdf5_path, 'r') as f:  # 修改6: 使用hdf5_path
            data_group = f['data']
            demo_count = data_group.attrs['demo_count']
        return f'demo_{demo_count:05d}'  # 返回当前演示的名称格式
    
    
    def get_mp4_save_path(self):  # 修改7: 增加demo_name和camera_name参数
        """
        生成视频保存路径。

        参数：
        - demo_name: 演示名称，如 'demo_00000'
        - camera_name: 相机名称，如 'camera_head_color'

        返回：
        - 视频文件的完整保存路径
        """
        # print("file_path:",self.file_path)
      #  print("uuid_dir 111111111111:",self.uuid_dir)

        # demo_dir = os.path.join(self.uuid_dir)  # 新增: 创建demo子目录
        demo_dir = os.path.join(self.camera_dir)
     #   print("Demo dir:",demo_dir)
        os.makedirs(demo_dir, exist_ok=True)  # 新增: 确保目录存在
        retpath = os.path.join(demo_dir) 
        #print("retpath:",retpath)
        return retpath  # 修改8: 返回demo子目录中的路径

    # def get_mp4_save_path(self) -> str:
    #     self.mp4_save_path = os.path.join(self.file_path.removesuffix('.hdf5'), self.get_cur_demo_name())
    #     return self.mp4_save_path
    
    def add_demo_data(self, demo_data, model_file=None):
        """
        添加一个新的演示（trajectory）。

        参数：
        - demo_data: 包含演示数据的字典，结构如下：
            {
                'states': np.ndarray (N, D),    # 机器人关节、夹爪、物体 的位姿、速度
                'actions': np.ndarray (N, A),   # 机械臂末端的位姿、夹爪的开合程度
                'objects': str,   # 物体的相关信息 json格式
                'goals': str     # 目标的相关信息 json格式
                'rewards': np.ndarray (N,),     # 奖励
                'dones': np.ndarray (N,),       # 完成标志
                'obs': dict of np.ndarrays      # 观测数据字典
                'timesteps': np.ndarray (N,)    # 仿真时间，单位为秒
                'language_instruction': str     # 语言指令（可选）
                'next_obs'                      # 如果未提供，将自动生成
                'camera_frames'                 # 可选，用于存储相机帧，格式为字典，键为相机名称，值为帧列表
                'camera_time_stamp': dict       # 相机时间戳，键为相机名称，值为时间戳列表
            }
        - model_file: MJCF MuJoCo 模型的 XML 字符串（可选，仅用于 robosuite 数据集）。
        """
        # 打开文件进行读写
        with h5py.File(self.hdf5_path, 'r+') as f:  # 修改9: 使用hdf5_path
            data_group = f['data']

            # 获取当前的演示计数和总样本数
            total_samples = data_group.attrs['total']
            # demo_name_h5 = self.get_cur_demo_name()   # 获取HDF5中的演示名称
            demo_name = self.get_cur_demo_name()
            demo_count = data_group.attrs['demo_count']  # 获取当前演示计数

            # demo_group = data_group.create_group(demo_name_h5)
            demo_group = data_group.create_group(demo_name)
            num_samples = demo_data['actions'].shape[0]
            demo_group.attrs['num_samples'] = num_samples

            if model_file:
                demo_group.attrs['model_file'] = model_file

            for key in ['states', 'actions', 'language_instruction', 'objects', 'goals', 'rewards', 'dones', 'timesteps', 'timestamps']:
                if key in demo_data:
                    demo_group.create_dataset(key, data=demo_data[key])
            
            # 处理task_info（任务信息）
            if 'task_info' in demo_data:
                task_info_group = demo_group.create_group('task_info')
                for task_key, task_value in demo_data['task_info'].items():
                    if isinstance(task_value, str):
                        # 字符串类型，转换为字节存储
                        task_info_group.create_dataset(task_key, data=task_value.encode('utf-8'))
                    else:
                        # 其他类型，直接存储
                        task_info_group.create_dataset(task_key, data=task_value)

            # 处理 obs
            obs_group = demo_group.create_group('obs')
            for obs_key, obs_data in demo_data['obs'].items():
                obs_group.create_dataset(obs_key, data=obs_data, compression="gzip", compression_opts=4)
 
            camera_group = demo_group.create_group('camera')
            for key in ['camera_frames', 'camera_time_stamp']:
                if key == 'camera_frames' and 'camera_frames' in demo_data:
                    # 如果存在 camera_frames，则创建相机帧数据集
                    demo_group.create_dataset('camera_frames', data=demo_data['camera_frames'], compression="gzip", compression_opts=4)
                if key == 'camera_time_stamp' and 'camera_time_stamp' in demo_data:
                    for camera_name, time_stamps in demo_data['camera_time_stamp'].items():
                        camera_group.create_dataset(camera_name, data=time_stamps, compression="gzip", compression_opts=4)

            # 自动生成 next_obs
            if 'next_obs' in demo_data:
                next_obs_data = demo_data['next_obs']
            else:
                next_obs_data = self._generate_next_obs(demo_data['obs'])

            next_obs_group = demo_group.create_group('next_obs')
            for obs_key, obs_data in next_obs_data.items():
                next_obs_group.create_dataset(obs_key, data=obs_data, compression="gzip", compression_opts=4)

            # 更新总样本数和演示计数
            total_samples += num_samples
            demo_count += 1
            data_group.attrs['total'] = total_samples
            data_group.attrs['demo_count'] = demo_count

    def save_single_camera_video(self, frames, video_path, fps):  # 新增: 单个相机视频保存方法
        """保存单个相机视频"""
        os.makedirs(os.path.dirname(video_path), exist_ok=True)
        print(f"Saving video at {video_path}")
        frame_height, frame_width = frames[0].shape[:2]
        isColor = (frames[0].ndim == 3)
        
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(video_path, fourcc, fps, (frame_width, frame_height), isColor=isColor)
        for frame in frames:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            out.write(frame)
        out.release()

    def _generate_next_obs(self, obs):
        """
        从 obs 自动生成 next_obs。

        参数：
        - obs: 观测数据字典

        返回：
        - next_obs: 下一个观测数据字典，结构与 obs 相同。
        """
        next_obs = {obs_key: [] for obs_key in obs.keys()}
        for obs_key, obs_data in obs.items():
            next_obs_data = obs_data[1:]
            next_obs_data = np.concatenate([next_obs_data, np.expand_dims(obs_data[-1], axis=0)], axis=0)
            next_obs[obs_key] = next_obs_data
        return next_obs

    def add_filter_key(self, filter_key_name, demo_names):
        with h5py.File(self.hdf5_path, 'r+') as f:  # 修改13: 使用hdf5_path
            mask_group = f['mask']
            # 将演示名称转换为字节串，适用于 HDF5
            demo_names_bytes = np.array([name.encode('utf-8') for name in demo_names], dtype='S')
            # 如果过滤键已存在，先删除再创建
            if filter_key_name in mask_group:
                del mask_group[filter_key_name]
            mask_group.create_dataset(filter_key_name, data=demo_names_bytes)
            print("Added filter key:", filter_key_name, "with", len(demo_names), "demos.")
            
    def remove_filter_key(self, filter_key_name):
        with h5py.File(self.hdf5_path, 'r+') as f:  # 修改14: 使用hdf5_path
            mask_group = f['mask']
            if filter_key_name in mask_group:
                del mask_group[filter_key_name]
                print("Removed filter key:", filter_key_name)

    def get_demo_names(self):
        with h5py.File(self.hdf5_path, 'r') as f:  # 修改15: 使用hdf5_path
            data_group = f['data']
            demo_names = [name for name in data_group.keys()]
        return demo_names
    
    def shuffle_demos(self, train_demo_ratio=0.8):
        """
        将 80% 的演示数据用于训练 (train)，剩余 20% 用于测试(valid)
        """

        if self.pathremoved:
            return
        self.remove_filter_key("train")
        self.remove_filter_key("valid")
        
        demo_names = self.get_demo_names()
        train_demos = []
        valid_demos = []
        for demo_name in demo_names:
            if np.random.rand() < train_demo_ratio:
                train_demos.append(demo_name)
                # print("Demo name: ", demo_name, " added to train.")
            else:
                valid_demos.append(demo_name)
                # print("Demo name: ", demo_name, " added to valid.")
                
        self.add_filter_key("train", train_demos)
        self.add_filter_key("valid", valid_demos)
            
    def finalize(self):
        """
        由于每次操作都已保存并关闭文件，因此此方法可为空或用于其他清理操作。
        """
        pass  # 在此示例中，无需执行任何操作
    
    # 删除原save_camera_video方法，使用新的save_single_camera_video方法

class DatasetReader:
    def __init__(self, file_path):
        """
        初始化 DatasetReader。

        参数：
        - file_path: 要读取的 HDF5 文件路径。
        """
        self.file_path = file_path

    def _get_env_args(self):
        """
        获取数据集的环境参数。

        返回：
        - env_args: 环境参数字典，结构如下：
            {
                'env_name': str,
                'type': str,
                'env_version': str,
                'env_kwargs': dict
            }
        """
        with h5py.File(self.file_path, 'r') as f:
            data_group = f['data']
            env_args = json.loads(data_group.attrs['env_args'])
        return env_args
    
    def get_env_name(self):
        """
        获取数据集的环境名称。

        返回：
        - env_name: 环境名称。
        """
        return self._get_env_args()['env_name']
    
    def get_env_version(self):
        """
        获取数据集的环境版本。

        返回：
        - env_version: 环境版本。
        """
        return self._get_env_args()['env_version']
    
    def get_env_kwargs(self):
        """
        获取数据集的环境参数。

        返回：
        - env_kwargs: 环境参数字典。
        """
        return self._get_env_args()['env_kwargs']
    
    def get_env_type(self):
        """
        获取数据集的环境类型。

        返回：
        - env_type: 环境类型。
        """
        return self._get_env_args()['type']

    def get_total_samples(self):
        """
        获取数据集中的总样本数。

        返回：
        - total_samples: 总样本数。
        """
        with h5py.File(self.file_path, 'r') as f:
            data_group = f['data']
            total_samples = data_group.attrs['total']
        return total_samples

    def get_demo_count(self):
        """
        获取数据集中的演示数量。

        返回：
        - demo_count: 演示数量。
        """
        with h5py.File(self.file_path, 'r') as f:
            data_group = f['data']
            demo_count = data_group.attrs['demo_count']
        return demo_count
    
    def get_demo_names(self, filter_key : str = None):
        """
        获取数据集中的所有演示名称。

        返回：
        - demo_names: 演示名称列表。
        """
        with h5py.File(self.file_path, 'r') as f:
            data_group = f['data']
            demo_names = [name for name in data_group.keys()]
            if filter_key is not None:
                mask_group = f['mask']
                filter_key_names = mask_group[filter_key]
                filter_key_names = [name.decode('utf-8') for name in filter_key_names]
                demo_names = [name for name in demo_names if name in filter_key_names]
        return demo_names
            
    def get_demo_data(self, demo_name):
        """
        获取指定名称的演示数据。

        参数：
        - demo_name: 演示名称。

        返回：
        - demo_data: 包含演示数据的字典，结构如下：
            {
                'states': np.ndarray (N, D),    # 机器人关节、夹爪、物体 的位姿、速度
                'actions': np.ndarray (N, A),   # 机械臂末端的位姿、夹爪的开合程度
                'objects': np.ndarray (N, O),   # 物体的位姿
                'goals': np.ndarray (N, G)      # 目标位姿（可选）
                'rewards': np.ndarray (N,),     # 奖励
                'dones': np.ndarray (N,),       # 完成标志
                'obs': dict of np.ndarrays      # 观测数据字典
                'timesteps': np.ndarray (N,)    # 仿真时间，单位为秒
                'language_instruction': str     # 语言指令（可选）
                'next_obs'                      # 如果未提供，将自动生成
                'camera_frames'                 # 用于存储相机帧 (可选)
            }
        """
        with h5py.File(self.file_path, 'r') as f:
            demo_group = f['data'][demo_name]
            demo_data = {
                'states': np.array(demo_group['states']),
                'actions': np.array(demo_group['actions']),
                'objects': np.array(demo_group['objects']) if 'objects' in demo_group else None,
                'goals': np.array(demo_group['goals']) if 'goals' in demo_group else None,
                'rewards': np.array(demo_group['rewards']),
                'dones': np.array(demo_group['dones']),
                'obs': {key: np.array(demo_group['obs'][key]) for key in demo_group['obs'].keys()},
                'timesteps': np.array(demo_group['timesteps']),
                'language_instruction': demo_group['language_instruction'][()] if 'language_instruction' in demo_group else None,
                'next_obs': {key: np.array(demo_group['next_obs'][key]) for key in demo_group['next_obs'].keys()},
                'camera_frames': np.array(demo_group['camera_frames'])
            }

            if 'timestamps' in demo_group:
                demo_data['timestamps'] = np.array(demo_group['timestamps'])
            if 'camera' in demo_group:
                demo_data['camera'] = {key: np.array(demo_group['camera'][key]) for key in demo_group['camera'].keys()}
        return demo_data



import abc


class Interpolator(object, metaclass=abc.ABCMeta):
    """
    General interpolator interface.
    """

    @abc.abstractmethod
    def get_interpolated_goal(self):
        """
        Provides the next step in interpolation given the remaining steps.

        Returns:
            np.array: Next interpolated step
        """
        raise NotImplementedError



import numpy as np

import orca_gym.adapters.robosuite.utils.transform_utils as T
from orca_gym.adapters.robosuite.controllers.interpolators.base_interpolator import Interpolator


class LinearInterpolator(Interpolator):
    """
    Simple class for implementing a linear interpolator.

    Abstracted to interpolate n-dimensions

    Args:
        ndim (int): Number of dimensions to interpolate

        controller_freq (float): Frequency (Hz) of the controller

        policy_freq (float): Frequency (Hz) of the policy model

        ramp_ratio (float): Percentage of interpolation timesteps across which we will interpolate to a goal position.

            :Note: Num total interpolation steps will be equal to np.floor(ramp_ratio * controller_freq / policy_freq)
                    i.e.: how many controller steps we get per action space update

        ori_interpolate (None or str): If set, assumes that we are interpolating angles (orientation)
            Specified string determines assumed type of input:

                `'euler'`: Euler orientation inputs
                `'quat'`: Quaternion inputs
    """

    def __init__(
        self,
        ndim,
        controller_freq,
        policy_freq,
        ramp_ratio=0.2,
        use_delta_goal=False,
        ori_interpolate=None,
    ):
        self.dim = ndim  # Number of dimensions to interpolate
        self.ori_interpolate = ori_interpolate  # Whether this is interpolating orientation or not
        self.order = 1  # Order of the interpolator (1 = linear)
        self.step = 0  # Current step of the interpolator
        self.total_steps = np.ceil(
            ramp_ratio * controller_freq / policy_freq
        )  # Total num steps per interpolator action
        self.use_delta_goal = use_delta_goal  # Whether to use delta or absolute goals (currently
        # not implemented yet- TODO)
        self.set_states(dim=ndim, ori=ori_interpolate)

    def set_states(self, dim=None, ori=None):
        """
        Updates self.dim and self.ori_interpolate.

        Initializes self.start and self.goal with correct dimensions.

        Args:
            ndim (None or int): Number of dimensions to interpolate

            ori_interpolate (None or str): If set, assumes that we are interpolating angles (orientation)
                Specified string determines assumed type of input:

                    `'euler'`: Euler orientation inputs
                    `'quat'`: Quaternion inputs
        """
        # Update self.dim and self.ori_interpolate
        self.dim = dim if dim is not None else self.dim
        self.ori_interpolate = ori if ori is not None else self.ori_interpolate

        # Set start and goal states
        if self.ori_interpolate is not None:
            if self.ori_interpolate == "euler":
                self.start = np.zeros(3)
            else:  # quaternions
                self.start = np.array((0, 0, 0, 1))
        else:
            self.start = np.zeros(self.dim)
        self.goal = np.array(self.start)

    def set_goal(self, goal):
        """
        Takes a requested (absolute) goal and updates internal parameters for next interpolation step

        Args:
            np.array: Requested goal (absolute value). Should be same dimension as self.dim
        """
        # First, check to make sure requested goal shape is the same as self.dim
        if goal.shape[0] != self.dim:
            print("Requested goal: {}".format(goal))
            raise ValueError(
                "LinearInterpolator: Input size wrong for goal; got {}, needs to be {}!".format(goal.shape[0], self.dim)
            )

        # Update start and goal
        self.start = np.array(self.goal)
        self.goal = np.array(goal)

        # Reset interpolation steps
        self.step = 0

    def get_interpolated_goal(self):
        """
        Provides the next step in interpolation given the remaining steps.

        NOTE: If this interpolator is for orientation, it is assumed to be receiving either euler angles or quaternions

        Returns:
            np.array: Next position in the interpolated trajectory
        """
        # Grab start position
        x = np.array(self.start)
        # Calculate the desired next step based on remaining interpolation steps
        if self.ori_interpolate is not None:
            # This is an orientation interpolation, so we interpolate linearly around a sphere instead
            goal = np.array(self.goal)
            if self.ori_interpolate == "euler":
                # this is assumed to be euler angles (x,y,z), so we need to first map to quat
                x = T.mat2quat(T.euler2mat(x))
                goal = T.mat2quat(T.euler2mat(self.goal))

            # Interpolate to the next sequence
            x_current = T.quat_slerp(x, goal, fraction=(self.step + 1) / self.total_steps)
            if self.ori_interpolate == "euler":
                # Map back to euler
                x_current = T.mat2euler(T.quat2mat(x_current))
        else:
            # This is a normal interpolation
            dx = (self.goal - x) / (self.total_steps - self.step)
            x_current = x + dx

        # Increment step if there's still steps remaining based on ramp ratio
        if self.step < self.total_steps - 1:
            self.step += 1

        # Return the new interpolated step
        return x_current



"""
This file contains the gym environment wrapper that is used
to provide a standardized environment API for training policies and interacting
with metadata present in datasets.
"""
import json
import numpy as np
from copy import deepcopy
from typing import Any, Dict, Optional, Tuple, Union

from orca_gym.environment import OrcaGymLocalEnv
from orca_gym.environment import OrcaGymLocalEnv
from orca_gym.adapters.robomimic.task.abstract_task import AbstractTask
import robomimic.envs.env_base as EB
import robomimic.utils.obs_utils as ObsUtils

import gymnasium as gym
from gymnasium import error, spaces
from gymnasium.spaces import Space


class RobomimicEnv(OrcaGymLocalEnv):
    metadata = {'render_modes': ['human', 'none'], 'version': '0.0.1', 'render_fps': 30}

    """
    Below we outline important methods that each EnvBase subclass needs to implement or override. 
    The implementation mostly follows the OpenAI-Gym convention.

    __init__(self, ...)

        Create the wrapped environment instance and assign it to self.env. 
        For example, in EnvGym it simply calls self.env = gym.make(env_name, **kwargs). 
        Refer to EnvRobosuite as an example of handling the constructor arguments.

    step(self, action)

        Take a step in the environment with an input action, return (observation, reward, done, info).

    reset(self)

        Reset the environment, return observation

    render(self, mode="human", height=None, width=None, camera_name=None, **kwargs)

        Render the environment if mode=='human'. 
        Return an RGB array if mode=='rgb_array'

    get_observation(self, obs=None)

        Return the current environment observation as a dictionary, unless obs is not None. 
        This function should process the raw environment observation to align with the input expected by the policy model. 
        For example, it should cast an image observation to float with value range 0-1 and shape format [C, H, W].

    is_success(self)

        Check if the task condition(s) is reached. 
        Should return a dictionary { str: bool } with at least a “task” key for the overall task success, 
        and additional optional keys corresponding to other task criteria.

    serialize(self)

        Aggregate and return all information needed to re-instantiate this environment in a dictionary. 
        This is the same as @env_meta - environment metadata stored in hdf5 datasets and used in robomimic/utils/env_utils.py.

    create_for_data_processing(cls, ...)

        (Optional) A class method that initializes an environment for data-postprocessing purposes, 
        which includes extracting observations, labeling dense / sparse rewards, and annotating dones in transitions. 
        This function should at least designate the list of observation modalities that are image / low-dimensional observations 
        by calling robomimic.utils.obs_utils.initialize_obs_utils_with_obs_specs().

    get_goal(self)

        (Optional) Get goal for a goal-conditional task

    set_goal(self, goal)

        (optional) Set goal with external specification

    get_state(self)

        (Optional) This function should return the underlying state of a simulated environment. 
        Should be compatible with reset_to.

    reset_to(self, state)

        (Optional) Reset to a specific simulator state. Useful for reproducing results.
    """


    def __init__(
        self,
        frame_skip: int,
        orcagym_addr: str,
        agent_names: list[str],
        time_step: float,        
        action_step: int,        
        camera_config: Dict[str, Any],
        **kwargs        
    ):
        self._action_step = action_step
        self._camera_config = camera_config
        self._task = None
        task = kwargs["task"]
        if isinstance(task, AbstractTask):
            self._task = task

        super().__init__(
            frame_skip = frame_skip,
            orcagym_addr = orcagym_addr,
            agent_names = agent_names,
            time_step = time_step,            
            **kwargs
        )

    def check_success(self):
        """
        Check if the task condition(s) is reached. Should return a dictionary
        { str: bool } with at least a "task" key for the overall task success,
        and additional optional keys corresponding to other task criteria.
        """        
        raise NotImplementedError
    
    def step(self, action) -> tuple:
        raise NotImplementedError
    
    def get_action_step(self):
        return self._action_step    
    
    def get_camera_config(self):
        return self._camera_config
    
    def get_env_version(self):
        """
        The dataset version must correspond to the version of the environment that was used to collect the data.
        """
        raise NotImplementedError
    
    def get_observation(self, obs=None):
        """
        Return the current environment observation as a dictionary, unless obs is not None.
        This function should process the raw environment observation to align with the input expected by the policy model.
        For example, it should cast an image observation to float with value range 0-1 and shape format [C, H, W].
        """
        raise NotImplementedError
    
    def normalize_action(self, action, min_action, max_action):
        """
        将原始动作归一化到 [-1, 1] 范围。
        
        :param action: 原始动作值
        :param min_action: 动作的最小值
        :param max_action: 动作的最大值
        :return: 归一化后的动作值
        """
        normalized_action = 2 * (action - min_action) / (max_action - min_action) - 1
        return np.clip(normalized_action, -1.0, 1.0)

    def denormalize_action(self, normalized_action, min_action, max_action):
        """
        将归一化的动作值反归一化回原始范围。
        
        :param normalized_action: 归一化后的动作值
        :param min_action: 动作的最小值
        :param max_action: 动作的最大值
        :return: 反归一化后的原始动作值
        """
        original_action = (normalized_action + 1) / 2 * (max_action - min_action) + min_action
        return original_action    


import abc  # for abstract base class definitions


class Device(metaclass=abc.ABCMeta):
    """
    Base class for all robot controllers.
    Defines basic interface for all controllers to adhere to.
    """

    @abc.abstractmethod
    def start_control(self):
        """
        Method that should be called externally before controller can
        start receiving commands.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_controller_state(self):
        """Returns the current state of the device, a dictionary of pos, orn, grasp, and reset."""
        raise NotImplementedError



from .controller_factory import controller_factory, load_controller_config, reset_controllers, get_pybullet_server
from .osc import OperationalSpaceController
from .joint_pos import JointPositionController
from .joint_vel import JointVelocityController
from .joint_tor import JointTorqueController


CONTROLLER_INFO = {
    "JOINT_VELOCITY": "Joint Velocity",
    "JOINT_TORQUE": "Joint Torque",
    "JOINT_POSITION": "Joint Position",
    "OSC_POSITION": "Operational Space Control (Position Only)",
    "OSC_POSE": "Operational Space Control (Position + Orientation)",
    "IK_POSE": "Inverse Kinematics Control (Position + Orientation) (Note: must have PyBullet installed)",
}

ALL_CONTROLLERS = CONTROLLER_INFO.keys()



import time
import warnings

import numpy as np

from orca_gym.scene.orca_gym_scene import Actor, MaterialInfo, LightInfo, OrcaGymScene

class SceneManager:
    def __init__(self, grpc_addr: str, init_env_callback = None):
        self._scene = OrcaGymScene(grpc_addr)
        self._init_env_callback = init_env_callback

    def register_init_env_callback(self, callback):
        self._init_env_callback = callback

    def add_actor(self, actor_name: str, asset_path: str, position: np.ndarray, rotation: np.ndarray, scale: float = 1.0):
        actor = Actor(actor_name, asset_path, position, rotation, scale)
        self._scene.add_actor(actor)

    def add_light(self, light_name: str, asset_path: str, position: np.ndarray, rotation: np.ndarray, scale: float = 1.0):
        actor= Actor(light_name, asset_path, position, rotation, 1.0)
        self._scene.add_actor(actor)

    def set_light_info(self, light_name: str, color = None, intensity = None):
        light_info = LightInfo(color, intensity)
        self._scene.set_light_info(light_name, light_info)

    def set_material_info(self, actor_name: str, base_color = None):
        material_info = MaterialInfo(base_color)
        self._scene.set_material_info(actor_name, material_info)

    def publish_scene(self):
        """
        Publish the scene to the ORCA Gym environment.
        """
        self._scene.publish_scene()
        if self._init_env_callback is None:
            warnings.warn("The ORCA Gym environment does not have an init_env_callback method.")

        time.sleep(3)
        self._init_env_callback()

    def publish_scene_without_init_env(self):
        '''
        只添加视觉效果，无需初始化环境。
        '''
        self._scene.publish_scene()

    def destory_scene(self):
        self._scene.publish_scene()
        self._scene.close()




"""
This module includes:

- Utility classes for modifying sim cameras

- Utility functions for performing common camera operations such as retrieving
camera matrices and transforming from world to camera frame or vice-versa.
"""
import json
import xml.etree.ElementTree as ET

import h5py
import numpy as np

import robosuite
import orca_gym.adapters.robosuite.utils.transform_utils as T
from orca_gym.adapters.robosuite.wrappers import DomainRandomizationWrapper, VisualizationWrapper


def get_camera_intrinsic_matrix(sim, camera_name, camera_height, camera_width):
    """
    Obtains camera intrinsic matrix.

    Args:
        sim (MjSim): simulator instance
        camera_name (str): name of camera
        camera_height (int): height of camera images in pixels
        camera_width (int): width of camera images in pixels
    Return:
        K (np.array): 3x3 camera matrix
    """
    cam_id = sim.model.camera_name2id(camera_name)
    fovy = sim.model.cam_fovy[cam_id]
    f = 0.5 * camera_height / np.tan(fovy * np.pi / 360)
    K = np.array([[f, 0, camera_width / 2], [0, f, camera_height / 2], [0, 0, 1]])
    return K


def get_camera_extrinsic_matrix(sim, camera_name):
    """
    Returns a 4x4 homogenous matrix corresponding to the camera pose in the
    world frame. MuJoCo has a weird convention for how it sets up the
    camera body axis, so we also apply a correction so that the x and y
    axis are along the camera view and the z axis points along the
    viewpoint.
    Normal camera convention: https://docs.opencv.org/2.4/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html

    Args:
        sim (MjSim): simulator instance
        camera_name (str): name of camera
    Return:
        R (np.array): 4x4 camera extrinsic matrix
    """
    cam_id = sim.model.camera_name2id(camera_name)
    camera_pos = sim.data.cam_xpos[cam_id]
    camera_rot = sim.data.cam_xmat[cam_id].reshape(3, 3)
    R = T.make_pose(camera_pos, camera_rot)

    # IMPORTANT! This is a correction so that the camera axis is set up along the viewpoint correctly.
    camera_axis_correction = np.array(
        [[1.0, 0.0, 0.0, 0.0], [0.0, -1.0, 0.0, 0.0], [0.0, 0.0, -1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]
    )
    R = R @ camera_axis_correction
    return R


def get_camera_transform_matrix(sim, camera_name, camera_height, camera_width):
    """
    Camera transform matrix to project from world coordinates to pixel coordinates.

    Args:
        sim (MjSim): simulator instance
        camera_name (str): name of camera
        camera_height (int): height of camera images in pixels
        camera_width (int): width of camera images in pixels
    Return:
        K (np.array): 4x4 camera matrix to project from world coordinates to pixel coordinates
    """
    R = get_camera_extrinsic_matrix(sim=sim, camera_name=camera_name)
    K = get_camera_intrinsic_matrix(
        sim=sim, camera_name=camera_name, camera_height=camera_height, camera_width=camera_width
    )
    K_exp = np.eye(4)
    K_exp[:3, :3] = K

    # Takes a point in world, transforms to camera frame, and then projects onto image plane.
    return K_exp @ T.pose_inv(R)


def get_camera_segmentation(sim, camera_name, camera_height, camera_width):
    """
    Obtains camera segmentation matrix.

    Args:
        sim (MjSim): simulator instance
        camera_name (str): name of camera
        camera_height (int): height of camera images in pixels
        camera_width (int): width of camera images in pixels
    Return:
        im (np.array): 2-channel segmented image where the first contains the
            geom types and the second contains the geom IDs
    """
    return sim.render(camera_name=camera_name, height=camera_height, width=camera_width, segmentation=True)[::-1]


def get_real_depth_map(sim, depth_map):
    """
    By default, MuJoCo will return a depth map that is normalized in [0, 1]. This
    helper function converts the map so that the entries correspond to actual distances.

    (see https://github.com/deepmind/dm_control/blob/master/dm_control/mujoco/engine.py#L742)

    Args:
        sim (MjSim): simulator instance
        depth_map (np.array): depth map with values normalized in [0, 1] (default depth map
            returned by MuJoCo)
    Return:
        depth_map (np.array): depth map that corresponds to actual distances
    """
    # Make sure that depth values are normalized
    assert np.all(depth_map >= 0.0) and np.all(depth_map <= 1.0)
    extent = sim.model.stat.extent
    far = sim.model.vis.map.zfar * extent
    near = sim.model.vis.map.znear * extent
    return near / (1.0 - depth_map * (1.0 - near / far))


def project_points_from_world_to_camera(points, world_to_camera_transform, camera_height, camera_width):
    """
    Helper function to project a batch of points in the world frame
    into camera pixels using the world to camera transformation.

    Args:
        points (np.array): 3D points in world frame to project onto camera pixel locations. Should
            be shape [..., 3].
        world_to_camera_transform (np.array): 4x4 Tensor to go from robot coordinates to pixel
            coordinates.
        camera_height (int): height of the camera image
        camera_width (int): width of the camera image

    Return:
        pixels (np.array): projected pixel indices of shape [..., 2]
    """
    assert points.shape[-1] == 3  # last dimension must be 3D
    assert len(world_to_camera_transform.shape) == 2
    assert world_to_camera_transform.shape[0] == 4 and world_to_camera_transform.shape[1] == 4

    # convert points to homogenous coordinates -> (px, py, pz, 1)
    ones_pad = np.ones(points.shape[:-1] + (1,))
    points = np.concatenate((points, ones_pad), axis=-1)  # shape [..., 4]

    # batch matrix multiplication of 4 x 4 matrix and 4 x 1 vectors to do robot frame to pixels transform
    mat_reshape = [1] * len(points.shape[:-1]) + [4, 4]
    cam_trans = world_to_camera_transform.reshape(mat_reshape)  # shape [..., 4, 4]
    pixels = np.matmul(cam_trans, points[..., None])[..., 0]  # shape [..., 4]

    # re-scaling from homogenous coordinates to recover pixel values
    # (x, y, z) -> (x / z, y / z)
    pixels = pixels / pixels[..., 2:3]
    pixels = pixels[..., :2].round().astype(int)  # shape [..., 2]

    # swap first and second coordinates to get pixel indices that correspond to (height, width)
    # and also clip pixels that are out of range of the camera image
    pixels = np.concatenate(
        (
            pixels[..., 1:2].clip(0, camera_height - 1),
            pixels[..., 0:1].clip(0, camera_width - 1),
        ),
        axis=-1,
    )

    return pixels


def transform_from_pixels_to_world(pixels, depth_map, camera_to_world_transform):
    """
    Helper function to take a batch of pixel locations and the corresponding depth image
    and transform these points from the camera frame to the world frame.

    Args:
        pixels (np.array): pixel coordinates of shape [..., 2]
        depth_map (np.array): depth images of shape [..., H, W, 1]
        camera_to_world_transform (np.array): 4x4 Tensor to go from pixel coordinates to world
            coordinates.

    Return:
        points (np.array): 3D points in robot frame of shape [..., 3]
    """

    # make sure leading dimensions are consistent
    pixels_leading_shape = pixels.shape[:-1]
    depth_map_leading_shape = depth_map.shape[:-3]
    assert depth_map_leading_shape == pixels_leading_shape

    # sample from the depth map using the pixel locations with bilinear sampling
    pixels = pixels.astype(float)
    im_h, im_w = depth_map.shape[-2:]
    depth_map_reshaped = depth_map.reshape(-1, im_h, im_w, 1)
    z = bilinear_interpolate(im=depth_map_reshaped, x=pixels[..., 1:2], y=pixels[..., 0:1])
    z = z.reshape(*depth_map_leading_shape, 1)  # shape [..., 1]

    # form 4D homogenous camera vector to transform - [x * z, y * z, z, 1]
    # (note that we need to swap the first 2 dimensions of pixels to go from pixel indices
    # to camera coordinates)
    cam_pts = [pixels[..., 1:2] * z, pixels[..., 0:1] * z, z, np.ones_like(z)]
    cam_pts = np.concatenate(cam_pts, axis=-1)  # shape [..., 4]

    # batch matrix multiplication of 4 x 4 matrix and 4 x 1 vectors to do camera to robot frame transform
    mat_reshape = [1] * len(cam_pts.shape[:-1]) + [4, 4]
    cam_trans = camera_to_world_transform.reshape(mat_reshape)  # shape [..., 4, 4]
    points = np.matmul(cam_trans, cam_pts[..., None])[..., 0]  # shape [..., 4]
    return points[..., :3]


def bilinear_interpolate(im, x, y):
    """
    Bilinear sampling for pixel coordinates x and y from source image im.
    Taken from https://stackoverflow.com/questions/12729228/simple-efficient-bilinear-interpolation-of-images-in-numpy-and-python
    """
    x = np.asarray(x)
    y = np.asarray(y)

    x0 = np.floor(x).astype(int)
    x1 = x0 + 1
    y0 = np.floor(y).astype(int)
    y1 = y0 + 1

    x0 = np.clip(x0, 0, im.shape[1] - 1)
    x1 = np.clip(x1, 0, im.shape[1] - 1)
    y0 = np.clip(y0, 0, im.shape[0] - 1)
    y1 = np.clip(y1, 0, im.shape[0] - 1)

    Ia = im[y0, x0]
    Ib = im[y1, x0]
    Ic = im[y0, x1]
    Id = im[y1, x1]

    wa = (x1 - x) * (y1 - y)
    wb = (x1 - x) * (y - y0)
    wc = (x - x0) * (y1 - y)
    wd = (x - x0) * (y - y0)

    return wa * Ia + wb * Ib + wc * Ic + wd * Id


class CameraMover:
    """
    A class for manipulating a camera.

    WARNING: This class will initially RE-INITIALIZE the environment.

    Args:
        env (MujocoEnv): Mujoco environment to modify camera
        camera (str): Which camera to mobilize during playback, e.g.: frontview, agentview, etc.
        init_camera_pos (None or 3-array): If specified, should be the (x,y,z) global cartesian pos to
            initialize camera to
        init_camera_quat (None or 4-array): If specified, should be the (x,y,z,w) global quaternion orientation to
            initialize camera to
    """

    def __init__(
        self,
        env,
        camera="frontview",
        init_camera_pos=None,
        init_camera_quat=None,
    ):
        # Store relevant values and initialize other values
        self.env = env
        self.camera = camera
        self.mover_body_name = f"{self.camera}_cameramover"

        # Get state
        state = self.env.sim.get_state().flatten()

        # Grab environment xml
        xml = env.sim.model.get_xml()

        # Modify xml to add mocap to move camera around
        xml = self.modify_xml_for_camera_movement(xml=xml, camera_name=self.camera)

        # Reset the environment and restore the state
        self.env.reset_from_xml_string(xml)
        self.env.sim.reset()
        self.env.sim.set_state_from_flattened(state)
        self.env.sim.forward()

        # Set initial camera pose
        self.set_camera_pose(pos=init_camera_pos, quat=init_camera_quat)

    def set_camera_pose(self, pos=None, quat=None):
        """
        Sets the camera pose, which optionally includes position and / or quaternion

        Args:
            pos (None or 3-array): If specified, should be the (x,y,z) global cartesian pos to set camera to
            quat (None or 4-array): If specified, should be the (x,y,z,w) global quaternion orientation to set camera to
        """
        if pos is not None:
            self.env.sim.data.set_mocap_pos(self.mover_body_name, pos)
        if quat is not None:
            self.env.sim.data.set_mocap_quat(self.mover_body_name, T.convert_quat(quat, to="wxyz"))

        # Make sure changes propagate in sim
        self.env.sim.forward()

    def get_camera_pose(self):
        """
        Grab the current camera pose, which optionally includes position and / or quaternion

        Returns:
            2-tuple:
                - 3-array: (x,y,z) camera global cartesian pos
                - 4-array: (x,y,z,w) camera global quaternion orientation
        """
        # Grab values from sim
        pos = self.env.sim.data.get_mocap_pos(self.mover_body_name)
        quat = T.convert_quat(self.env.sim.data.get_mocap_quat(self.mover_body_name), to="xyzw")

        return pos, quat

    def modify_xml_for_camera_movement(self, xml, camera_name):
        """
        Cameras in mujoco are 'fixed', so they can't be moved by default.
        Although it's possible to hack position movement, rotation movement
        does not work. An alternative is to attach a camera to a mocap body,
        and move the mocap body.

        This function modifies the camera with name @camera_name in the xml
        by attaching it to a mocap body that can move around freely. In this
        way, we can move the camera by moving the mocap body.

        See http://www.mujoco.org/forum/index.php?threads/move-camera.2201/ for
        further details.

        Args:
            xml (str): Mujoco sim XML file as a string
            camera_name (str): Name of camera to tune
        """
        tree = ET.fromstring(xml)
        wb = tree.find("worldbody")

        # find the correct camera
        camera_elem = None
        cameras = wb.findall("camera")
        for camera in cameras:
            if camera.get("name") == camera_name:
                camera_elem = camera
                break
        assert camera_elem is not None

        # add mocap body
        mocap = ET.SubElement(wb, "body")
        mocap.set("name", self.mover_body_name)
        mocap.set("mocap", "true")
        mocap.set("pos", camera.get("pos"))
        mocap.set("quat", camera.get("quat"))
        new_camera = ET.SubElement(mocap, "camera")
        new_camera.set("mode", "fixed")
        new_camera.set("name", camera.get("name"))
        new_camera.set("pos", "0 0 0")

        # remove old camera element
        wb.remove(camera_elem)

        return ET.tostring(tree, encoding="utf8").decode("utf8")

    def rotate_camera(self, point, axis, angle):
        """
        Rotate the camera view about a direction (in the camera frame).

        Args:
            point (None or 3-array): (x,y,z) cartesian coordinates about which to rotate camera in camera frame. If None,
                assumes the point is the current location of the camera
            axis (3-array): (ax,ay,az) axis about which to rotate camera in camera frame
            angle (float): how much to rotate about that direction

        Returns:
            2-tuple:
                pos: (x,y,z) updated camera position
                quat: (x,y,z,w) updated camera quaternion orientation
        """
        # current camera rotation + pos
        camera_pos = np.array(self.env.sim.data.get_mocap_pos(self.mover_body_name))
        camera_rot = T.quat2mat(T.convert_quat(self.env.sim.data.get_mocap_quat(self.mover_body_name), to="xyzw"))

        # rotate by angle and direction to get new camera rotation
        rad = np.pi * angle / 180.0
        R = T.rotation_matrix(rad, axis, point=point)
        camera_pose = np.zeros((4, 4))
        camera_pose[:3, :3] = camera_rot
        camera_pose[:3, 3] = camera_pos
        camera_pose = camera_pose @ R

        # Update camera pose
        pos, quat = camera_pose[:3, 3], T.mat2quat(camera_pose[:3, :3])
        self.set_camera_pose(pos=pos, quat=quat)

        return pos, quat

    def move_camera(self, direction, scale):
        """
        Move the camera view along a direction (in the camera frame).

        Args:
            direction (3-array): direction vector for where to move camera in camera frame
            scale (float): how much to move along that direction
        """
        # current camera rotation + pos
        camera_pos = np.array(self.env.sim.data.get_mocap_pos(self.mover_body_name))
        camera_quat = self.env.sim.data.get_mocap_quat(self.mover_body_name)
        camera_rot = T.quat2mat(T.convert_quat(camera_quat, to="xyzw"))

        # move along camera frame axis and set new position
        camera_pos += scale * camera_rot.dot(direction)
        self.set_camera_pose(pos=camera_pos)

        return camera_pos, camera_quat


class DemoPlaybackCameraMover(CameraMover):
    """
    A class for playing back demonstrations and recording the resulting frames with the flexibility of a mobile camera
    that can be set manually or panned automatically frame-by-frame

    Note: domain randomization is also supported for playback!

    Args:
        demo (str): absolute fpath to .hdf5 demo
        env_config (None or dict): (optional) values to override inferred environment information from demonstration.
            (e.g.: camera h / w, depths, segmentations, etc...)
            Any value not specified will be inferred from the extracted demonstration metadata
            Note that there are some specific arguments that MUST be set a certain way, if any of these values
            are specified with @env_config, an error will be raised
        replay_from_actions (bool): If True, will replay demonstration's actions. Otherwise, replays will be hardcoded
            from the demonstration states
        visualize_sites (bool): If True, will visualize sites during playback. Note that this CANNOT be paired
            simultaneously with camera segmentations
        camera (str): Which camera to mobilize during playback, e.g.: frontview, agentview, etc.
        init_camera_pos (None or 3-array): If specified, should be the (x,y,z) global cartesian pos to
            initialize camera to
        init_camera_quat (None or 4-array): If specified, should be the (x,y,z,w) global quaternion orientation to
            initialize camera to
        use_dr (bool): If True, will use domain randomization during playback
        dr_args (None or dict): If specified, will set the domain randomization wrapper arguments if using dr
    """

    def __init__(
        self,
        demo,
        env_config=None,
        replay_from_actions=False,
        visualize_sites=False,
        camera="frontview",
        init_camera_pos=None,
        init_camera_quat=None,
        use_dr=False,
        dr_args=None,
    ):
        # Store relevant values and initialize other values
        self.camera_id = None
        self.replay_from_actions = replay_from_actions
        self.states = None
        self.actions = None
        self.step = None
        self.n_steps = None
        self.current_ep = None
        self.started = False

        # Load the demo
        self.f = h5py.File(demo, "r")

        # Extract relevant info
        env_info = json.loads(self.f["data"].attrs["env_info"])

        # Construct default env arguments
        default_args = {
            "has_renderer": False,
            "has_offscreen_renderer": True,
            "ignore_done": True,
            "use_camera_obs": True,
            "reward_shaping": True,
            "hard_reset": False,
            "camera_names": camera,
        }

        # If custom env_config is specified, make sure that there's no overlap with default args and merge with config
        if env_config is not None:
            for k in env_config.keys():
                assert k not in default_args, f"Key {k} cannot be specified in env_config!"
            env_info.update(env_config)

        # Merge in default args
        env_info.update(default_args)

        # Create env
        env = orca_gym.adapters.robosuite.make(**env_info)

        # Optionally wrap with visualization wrapper
        if visualize_sites:
            env = VisualizationWrapper(env=self.env)

        # Optionally use domain randomization if specified
        self.use_dr = use_dr
        if self.use_dr:
            default_dr_args = {
                "seed": 1,
                "randomize_camera": False,
                "randomize_every_n_steps": 10,
            }
            default_dr_args.update(dr_args)
            env = DomainRandomizationWrapper(
                env=self.env,
                **default_dr_args,
            )

        # list of all demonstrations episodes
        self.demos = list(self.f["data"].keys())

        # Run super init
        super().__init__(
            env=env,
            camera=camera,
            init_camera_pos=init_camera_pos,
            init_camera_quat=init_camera_quat,
        )

        # Load episode 0 by default
        self.load_episode_xml(demo_num=0)

    def load_episode_xml(self, demo_num):
        """
        Loads demo episode with specified @demo_num into the simulator.

        Args:
            demo_num (int): Demonstration number to load
        """
        # Grab raw xml file
        ep = self.demos[demo_num]
        model_xml = self.f[f"data/{ep}"].attrs["model_file"]

        # Reset environment
        self.env.reset()
        xml = self.env.edit_model_xml(model_xml)
        xml = self.modify_xml_for_camera_movement(xml, camera_name=self.camera)
        self.env.reset_from_xml_string(xml)
        self.env.sim.reset()

        # Update camera info
        self.camera_id = self.env.sim.model.camera_name2id(self.camera)

        # Load states and actions
        self.states = self.f[f"data/{ep}/states"].value
        self.actions = np.array(self.f[f"data/{ep}/actions"].value)

        # Set initial state
        self.env.sim.set_state_from_flattened(self.states[0])

        # Reset step count and set current episode number
        self.step = 0
        self.n_steps = len(self.actions)
        self.current_ep = demo_num

        # Notify user of loaded episode
        print(f"Loaded episode {demo_num}.")

    def grab_next_frame(self):
        """
        Grabs the next frame in the demo sequence by stepping the simulation and returning the resulting value(s)

        Returns:
            dict: Keyword-mapped np.arrays from the demonstration sequence, corresponding to all image modalities used
                in the playback environment (e.g.: "image", "depth", "segmentation_instance")
        """
        # Make sure the episode isn't completed yet, if so, we load the next episode
        if self.step == self.n_steps:
            self.load_episode_xml(demo_num=self.current_ep + 1)

        # Step the environment and grab obs
        if self.replay_from_actions:
            obs, _, _, _ = self.env.step(self.actions[self.step])
        else:  # replay from states
            self.env.sim.set_state_from_flattened(self.states[self.step + 1])
            if self.use_dr:
                self.env.step_randomization()
            self.env.sim.forward()
            obs = self.env._get_observation()

        # Increment the step counter
        self.step += 1

        # Return all relevant frames
        return {k.split(f"{self.camera}_")[-1]: obs[k] for k in obs if self.camera in k}

    def grab_episode_frames(self, demo_num, pan_point=(0, 0, 0.8), pan_axis=(0, 0, 1), pan_rate=0.01):
        """
        Playback entire episode @demo_num, while optionally rotating the camera about point @pan_point and
            axis @pan_axis if @pan_rate > 0

        Args:
            demo_num (int): Demonstration episode number to load for playback
            pan_point (3-array): (x,y,z) cartesian coordinates about which to rotate camera in camera frame
            pan_direction (3-array): (ax,ay,az) axis about which to rotate camera in camera frame
            pan_rate (float): how quickly to pan camera if not 0

        Returns:
            dict: Keyword-mapped stacked np.arrays from the demonstration sequence, corresponding to all image
                modalities used in the playback environment (e.g.: "image", "depth", "segmentation_instance")

        """
        # First, load env
        self.load_episode_xml(demo_num=demo_num)

        # Initialize dict to return
        obs = self.env._get_observation()
        frames_dict = {k.split(f"{self.camera}_")[-1]: [] for k in obs if self.camera in k}

        # Continue to loop playback steps while there are still frames left in the episode
        while self.step < self.n_steps:
            # Take playback step and add frames
            for k, frame in self.grab_next_frame().items():
                frames_dict[k].append(frame)

            # Update camera pose
            self.rotate_camera(point=pan_point, axis=pan_axis, angle=pan_rate)

        # Stack all frames and return
        return {k: np.stack(frames) for k, frames in frames_dict.items()}



from colorama import Fore, Style
from orca_gym.environment import OrcaGymLocalEnv
from orca_gym.adapters.robomimic.task.abstract_task import AbstractTask
import random


import json
import re

import numpy as np
from colorama import Fore, Style
from orca_gym.environment import OrcaGymLocalEnv
from orca_gym.adapters.robomimic.task.abstract_task import AbstractTask
import random

class PickPlaceTask(AbstractTask):
    def __init__(self, config: dict):
        
        if config.__len__():
            super().__init__(config)
        else:
            super().__init__()

    def get_task(self, env: OrcaGymLocalEnv):
        # 随机灯光说明是增广任务
        if self.random_light:
            self._get_augmentation_task_(env, self.data, sample_range=self.sample_range)
        else:
            self._get_teleperation_task_(env)

        print(
            f"{Fore.WHITE}level: {self.level_name}{Style.RESET_ALL}  "
            f"object: {Fore.CYAN}{Style.BRIGHT}{self.target_object}{Style.RESET_ALL}  to  "
            f"goal:   {Fore.MAGENTA}{Style.BRIGHT}{self.goal_bodys[0]}{Style.RESET_ALL}"
            )
    def _get_teleperation_task_(self, env: OrcaGymLocalEnv) -> dict[str, str]:
        """
        随机选一个 object_bodys 里的物体，配对到第一个 goal_bodys。
        """
        self.generate_object(env, 3, 5)
        while True:
            # 随机摆放
            self.random_objs_and_goals(env, random_rotation=True)

            # 如果没有可用的物体或目标，直接返回空
            if not self.object_bodys or not self.goal_bodys:
                return None

            # 从 object_bodys 随机选一个
            self.target_object = random.choice(self.object_bodys)
            print(f"self.target_object: {self.target_object}")
            # 只取第一个 goal
            goal_name = self.goal_bodys[0]

            # 记录到 task_dict 并返回

            objs = self.randomized_object_positions
            goal_body = self.goal_bodys[0]

            return self.target_object

    def _get_augmentation_task_(self, env: OrcaGymLocalEnv, data: dict, sample_range = 0.0) ->dict[str, str]:
        '''
        获取一个增广任务
        '''
        self._restore_objects_(env, data['objects'])
        self._set_target_object_(env, data)
        if sample_range > 0.0:
            self.resample_objects(env, sample_range)
        self.__random_count__ += 1


    def _resample_objects_(self, env: OrcaGymLocalEnv, data: dict, sample_range: float = 0.0):
        target_obj_joint_name = env.joint(self.target_object + "_joint")
        target_obj_position = np.array([0.0, 0.0, 0.0], dtype=np.float32)
        objects = data.get("objects", [])
        for obj in objects:
            joint_name = obj["joint_name"].decode('utf-8') if isinstance(obj["joint_name"], (bytes, bytearray)) else obj["joint_name"]
            if target_obj_joint_name == joint_name:
                target_obj_position = np.array(obj["position"], dtype=np.float32)
                break

        resample_success = False
        for i in range(100):
            self.random_objs_and_goals(env, False, target_obj_joint_name)
            target_obj_joint_qpos = env.query_joint_qpos([target_obj_joint_name])[target_obj_joint_name]
            target_obj_position_delta = np.linalg.norm(target_obj_joint_qpos[:2] - target_obj_position[:2])
            if target_obj_position_delta < sample_range:
                resample_success = True
                break

    def process_goals(self, env:OrcaGymLocalEnv):
        """
        处理目标（goals）的信息，返回目标的bounding box（最大最小坐标）。
        :param goal_positions: 目标位置字典
        :return: 目标信息的条目，目标位置数组，和bounding box数据
        """
        goal_joints = [env.joint(jn) for jn in self.goal_joints]
        goal_positions = env.query_joint_qpos(goal_joints)

        goal_entries = []
        goal_positions_list = []
        goal_bounding_boxes = {}  # 用于存储目标的bounding box信息

        for goal_joint_name, qpos in goal_positions.items():
            # 获取目标的尺寸
            goal_name = goal_joint_name.replace("_joint", "")
            info = env.get_goal_bounding_box(goal_name)

            # 如果没有尺寸信息，跳过目标
            if not info:
                print(f"Error: No geometry size information found for goal {goal_name}")
                continue

            mn = np.array(info["min"]).flatten()
            mx = np.array(info["max"]).flatten()
            sz = mx - mn

            # 添加目标位置信息
            goal_entries.append({
                "joint_name": goal_name,
                "position": qpos[:3].tolist(),
                "orientation": qpos[3:].tolist(),
                "min": mn.tolist(),
                "max": mx.tolist(),
                "size": sz.tolist()
            })

            goal_positions_list.append(qpos[:3])  # 仅记录目标位置

        goal_positions_array = np.array(goal_positions_list)

        # 返回目标数据及bounding box信息
        return goal_entries, goal_positions_array

    def is_success(self, env: OrcaGymLocalEnv):
        target_pos, _, _ = env.get_body_xpos_xmat_xquat([env.body(self.target_object)])
        pos_vec = target_pos[0] if hasattr(target_pos, 'ndim') and target_pos.ndim > 1 else target_pos
        xy = pos_vec[:2]

        goal_entries, goal_positions = self.process_goals(env)
        #只有一个goal, 只取第一个
        goal_entry, goal_position = goal_entries[0], goal_positions[0]
        # 获取目标区域边界
        gmin = goal_entry['min'][:2]
        gmax = goal_entry['max'][:2]
        
        if xy[0] < gmin[0] or xy[0] > gmax[0] or xy[1] < gmin[1] or xy[1] > gmax[1]:
            return False
        return True

    def get_language_instruction(self) -> str:
        if not self.target_object:
            return "Do something."
        obj_str = "object: " + self.target_object
        goal_str = "goal: " + self.goal_bodys[0]

        return f"level: {self.level_name}  {obj_str} to {goal_str}"
        # return f"In the level {self.level_name} put the {self.target_object} into the {self.goal_bodys[0]}."
        # return f"In the {self.level_name} scene, put the {self.target_object} into the {self.goal_bodys[0]}."


class TaskStatus:
    """
    Enum class for task status
    """
    NOT_STARTED = "not_started"
    GET_READY = "get_ready"
    BEGIN = "begin"
    SUCCESS = "success"
    FAILURE = "failure"
    END = "end"



from .subproc_vec_env import OrcaGymAsyncSubprocVecEnv
from .orca_gym_async_env import OrcaGymAsyncEnv
from .orca_gym_async_agent import OrcaGymAsyncAgent

__all__ = ['OrcaGymAsyncSubprocVecEnv', 'OrcaGymAsyncEnv', 'OrcaGymAsyncAgent']


import sys

from PyQt5.QtGui import QPalette, QColor
from PyQt5.QtWidgets import QFrame, QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QDesktopWidget

import videoPlayer
from orca_gym.hdf5_viewer.videoPlayer import VideoPlayer


class HDF5Viewer(QMainWindow):
    def __init__(self):
        super().__init__()
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("HDF5/MP4 Viewer")
        self.setGeometry(100, 100, 1600, 900)

        main_widget = QWidget()
        main_layout = QHBoxLayout()

        left_frame = VideoPlayer() #视频区域
        right_frame = QWidget() #数据区域

        #设置背景颜色为黑色
        left_palette = left_frame.palette()
        left_palette.setColor(QPalette.Window, QColor(55, 55, 55))
        left_frame.setPalette(left_palette)
        left_frame.setAutoFillBackground(True)

        divider_line = QFrame()
        divider_line.setFrameShape(QFrame.VLine)  # 垂直分割线
        divider_line.setFrameShadow(QFrame.Sunken)
        divider_line.setLineWidth(1)  # 1像素宽的分割线
        divider_line.setMidLineWidth(0)
        divider_line.setStyleSheet("background-color: #a0a0a0;")

        main_layout.addWidget(left_frame, 3)
        main_layout.addWidget(divider_line)
        main_layout.addWidget(right_frame, 1)

        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)

        self.center()
        self.show()

    def center(self):
        screen_geometry = QDesktopWidget().availableGeometry()
        screen_center = screen_geometry.center()

        window_geometry = self.frameGeometry()
        window_geometry.moveCenter(screen_center)

        self.move(window_geometry.topLeft())

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = HDF5Viewer()
    sys.exit(app.exec_())


import json
import re

from orca_gym.environment import OrcaGymLocalEnv
from orca_gym.utils import rotations
from orca_gym.adapters.robomimic.task.scene_manage import SceneManager

from orca_gym.scene.orca_gym_scene import Actor, MaterialInfo, LightInfo, OrcaGymScene
import numpy as np
import random
import warnings, time

DEFAULT_CONFIG = {
    "random_object": True, # 是否随机Object的位置，object应该在Agent下面
    "type": "pick_and_place", # 任务类型，默认是pick_and_place
    "object_bodys": [],
    "object_sites": [],
    "object_joints": [],

    "random_goal": False, # 是否随机Goal的位置，goal应该在Agent下面
    "goal_bodys": [],
    "goal_sites": [],
    "goal_joints": [],

    "level_name": None,
    "random_cycle": 20,
    "range": [0.0, 0.0, 0.0, 0.0],
    "random_actor": False, # 是否随机添加Actor，actor与Agent同级
    "random_actor_position": False, # 是否随机Actor的位置
    "random_actor_rotation": False, # 是否随机Actor的旋转
    "random_actor_color": False, # 是否随机Actor的颜色
    "actors": [], # actor的命名
    "actors_spawnable": [], # actor的spawnable name, spawnable在Asset/Prefabs下面
    "center": [0, 0, 0], # actor的中心位置
    "infinity": [1000, 1000, 1], # actor的无限远位置
    "bound": [[-1, 1], [-1, 1], [0, 2]], # 以center点为中心，距离中心的边界位置
    "description": [],

    "random_light": False, # 是否随机添加光源
    "random_light_position": False, # 是否随机光源的位置
    "random_light_rotation": False, # 是否随机光源的旋转
    "lights": [], # 光源的命名
    "lights_spawnable": [], # 光源的spawnable name, spawnable在Asset/Prefabs下面
    "light_center": [0, 0, 0], # 光源的中心位置
    "light_bound": [[-1, 1], [-1, 1], [0, 2]], # 以center点为中心，距离中心的边界位置
    "light_description": [],
    "num_lights": 1,
    "grpc_addr": "localhost:50051"
}

class TaskStatus:
    """
    Enum class for task status
    """
    NOT_STARTED = "not_started"
    GET_READY = "get_ready"
    BEGIN = "begin"
    SUCCESS = "success"
    FAILURE = "failure"
    END = "end"


class AbstractTask:
    """
    :param prompt: The prompt for the task.
    :param config: The configuration for the task.
    """
    def __init__(self, config: dict = DEFAULT_CONFIG, init_env_callback = None):
        self.object_bodys = []
        self.object_sites = []
        self.object_joints = []
        self.goal_bodys = []
        self.goal_sites = []
        self.goal_joints = []
        self.randomized_object_positions = []
        self.level_name = None
        self.type = "pick_and_place"  # 任务类型，默认是pick_and_place
        self.range = {}
        self.random_actor = False
        self.__random_count__ = 0 # 用于随机Actor的计数， 每次随机摆放Actor和灯光都要+1
        self.random_cycle = 20
        self.random_actor_position = False
        self.random_actor_rotation = False
        self.actors = []
        self.actors_spawnable = []
        self.center = []
        self.bound = []
        self.description = []
        self.grpc_addr = None

        self.random_light = False
        self.random_light_position = False
        self.random_actor_rotation = False
        self.lights = []
        self.lights_spawnable = []
        self.light_center = []
        self.light_bound = []
        self.light_description = []
        self.load_config(config)

        self.scene = SceneManager(grpc_addr=self.grpc_addr, init_env_callback=init_env_callback)

        self.first_spawn = True
        self.data = {} #来着增广的数据
        self.sample_range = 0.0 #增广采样范围
        self.target_object = None


    def get_object_sites(self):
        return self.object_sites

    def get_goal_sites(self):
        return self.goal_sites

    def get_object_bodys(self):
        return self.object_bodys

    def get_goal_bodys(self):
        return self.goal_bodys

    def set_objects(self, objects: list[str]):
        self.object_bodys = objects

    def set_goals(self, goals: list[str]):
        self.goal_bodys = goals

    def set_object_sites(self, object_sites: list[str]):
        self.object_sites = object_sites

    def set_goal_sites(self, goal_sites: list[str]):
        self.goal_sites = goal_sites

    def get_object_joints_xpos(self, env: OrcaGymLocalEnv):
        env_object_joints = [env.joint(joint_name) for joint_name in self.object_joints]
        return env.query_joint_qpos(env_object_joints)

    def get_goal_joints_xpos(self, env: OrcaGymLocalEnv):
        env_goal_joints = [env.joint(joint_name) for joint_name in self.goal_joints]
        return env.query_joint_qpos(env_goal_joints)

    def get_object_xpos_xmat_xquat(self, env: OrcaGymLocalEnv):
        env_objects_body = [env.body(body_name) for body_name in self.object_bodys]
        return env.get_body_xpos_xmat_xquat(env_objects_body)

    def get_goal_xpos_xmat_xquat(self, env: OrcaGymLocalEnv):
        env_goals_body = [env.body(body_name) for body_name in self.goal_bodys]
        return env.get_body_xpos_xmat_xquat(env_goals_body)

    def generate_object(self, env:OrcaGymLocalEnv, pick_min, pick_max):
        '''
        从场景中随机挑选actors里的几个
        '''
        if self.random_actor:
            if self.__random_count__ % self.random_cycle == 0:
                # 将object_bodys放回到无限远处
                pos = self.infinity + [1, 0, 0, 0]
                qpos_dict = {env.joint(joint_name): pos for joint_name in self.object_joints}
                env.set_joint_qpos(qpos_dict)

                n_select  = random.randint(pick_min, pick_max)
                idxs      = random.sample(range(len(self.actors)), k=n_select)
                self.object_bodys = [self.actors[idx] for idx in idxs ]
                self.object_sites = [f"{self.actors[idx]}site" for idx in idxs]
                self.object_joints = [f"{self.actors[idx]}_joint" for idx in idxs]

            self.__random_count__ += 1

    def load_config(self, config: dict):
        """
        Load task configuration from a dictionary.
        """
        if config is None or len(config) == 0:
            return
        self.object_bodys = self.__get_config_setting__("object_bodys", config)
        self.object_sites = self.__get_config_setting__("object_sites", config)
        self.object_joints = self.__get_config_setting__("object_joints", config)
        self.goal_bodys = self.__get_config_setting__("goal_bodys", config)
        self.goal_sites = self.__get_config_setting__("goal_sites", config)
        self.goal_joints = self.__get_config_setting__("goal_joints", config)

        self.level_name = self.__get_config_setting__("level_name", config)
        self.range = self.__get_config_setting__("range", config)
        self.type = self.__get_config_setting__("type", config)
        self.random_cycle = self.__get_config_setting__("random_cycle", config)
        self.random_actor = self.__get_config_setting__("random_actor", config)
        self.random_actor_position = self.__get_config_setting__("random_actor_position", config)
        self.random_actor_rotation = self.__get_config_setting__("random_actor_rotation", config)
        self.random_actor_color = self.__get_config_setting__("random_actor_color", config)
        self.center = self.__get_config_setting__("center", config)
        self.infinity = self.__get_config_setting__("infinity", config)
        self.bound = self.__get_config_setting__("bound", config)
        self.actors = self.__get_config_setting__("actors", config)
        self.actors_spawnable = self.__get_config_setting__("actors_spawnable", config)
        self.actor_description = self.__get_config_setting__("description", config)

        self.lights = self.__get_config_setting__("lights", config)
        self.lights_spawnable = self.__get_config_setting__("lights_spawnable", config)
        self.light_center = self.__get_config_setting__("light_center", config)
        self.light_bound = self.__get_config_setting__("light_bound", config)
        self.random_light = self.__get_config_setting__("random_light", config)
        self.random_light_position = self.__get_config_setting__("random_light_position", config)
        self.random_light_rotation = self.__get_config_setting__("random_light_rotation", config)
        self.light_description = self.__get_config_setting__("light_description", config)
        self.num_lights = self.__get_config_setting__("num_lights", config)
        self.grpc_addr = self.__get_config_setting__("grpc_addr", config)

    def spawn_scene(self, env: OrcaGymLocalEnv):
        if self.first_spawn and not self.random_light:
            self.scene.publish_scene_without_init_env()
            self.generate_actors()

            self.scene.publish_scene()
            self.first_spawn = False
        elif self.random_light:
        # 周期性随机添加灯光
            if self.random_light and self.__random_count__ % self.random_cycle == 0:
                self.scene.publish_scene_without_init_env()
                self.generate_actors()
                light_idxs = self.generate_lights()
                self.scene.publish_scene()

                for idx in light_idxs:
                    self.set_light_info(self.lights[idx])
                for actor in self.actors:
                    self.set_actor_material(actor)

    #todo: 需要重写这部分代码， 这里应该只需要做随机位置，位置是否合理应该由具体任务决定
    def random_objs_and_goals(self, env: OrcaGymLocalEnv, random_rotation=True, target_obj_joint_name=None):
        object_bodys = [env.body(bn) for bn in self.object_bodys]
        obj_joints  = [env.joint(jn) for jn in self.object_joints]
        goal_joints = [env.joint(jn) for jn in self.goal_joints]

        dummy = env.site("dummy_site")
        info  = env.query_site_pos_and_quat([dummy])[dummy]
        base_pos, base_quat = info["xpos"], info["xquat"]

        def _get_qpos_not_in_goal_range(goal0_pos, base_pos, base_quat, joint):
            if target_obj_joint_name is not None:
                target_obj_pos = env.query_joint_qpos([target_obj_joint_name])[target_obj_joint_name][:3]
            else:
                target_obj_pos = None

            while True:
                lx = np.random.uniform(*self.range["x"])
                ly = np.random.uniform(*self.range["y"])
                lz = self.range["z"]
                lr = self.range["r"]
                local_pos = np.array([lx, ly, lz], dtype=np.float32)

                world_pos = rotations.quat_rot_vec(base_quat, local_pos) + base_pos
                yaw = np.random.uniform(np.pi * 0.25, np.pi * 0.75)
                world_quat = rotations.euler2quat([0.0, 0.0, yaw])
                obj_qpos = np.concatenate([world_pos, world_quat])

                if np.linalg.norm(world_pos[:2] - goal0_pos[:2]) > lr:
                    if target_obj_pos is not None and target_obj_joint_name != joint:
                        if np.linalg.norm(world_pos[:2] - target_obj_pos[:2]) < lr:
                            # print(f"Too close between {joint} and {target_obj_joint_name}, distance: {np.linalg.norm(world_pos[:2] - target_obj_pos[:2])}")
                            continue
                    break
            
            return obj_qpos

        def _find_obj_place_no_contact(find_target_obj=False):
            placed = []
            goal0_pos = env.query_joint_qpos(goal_joints)[goal_joints[0]][:3]
            placed_body = []
            for joint, body in zip(obj_joints, object_bodys):
                if find_target_obj and joint != target_obj_joint_name:
                    continue

                # 跳过落进目标的
                obj_qpos = _get_qpos_not_in_goal_range(goal0_pos, base_pos, base_quat, joint)
                # print("[Debug] Placing object on joint:", joint, "at position:", obj_qpos)
                if not random_rotation:
                    org_qpos = env.query_joint_qpos([joint])[joint]
                    obj_qpos[3:] = org_qpos[3:]
                env.set_joint_qpos({joint: obj_qpos})
                env.mj_forward()
                placed_body.append(body)

                # 跳过有碰撞的
                contacts = env.query_contact_simple()
                find_contact = False
                for contact in contacts:
                    body1 = env.model.get_geom_body_name(contact["Geom1"])
                    body2 = env.model.get_geom_body_name(contact["Geom2"])
                    if body1 in placed_body and body2 in placed_body:
                        # print(f"[Debug] Found contact between {body1} and {body2}")
                        find_contact = True
                        break
                if find_contact:
                    return None

                placed.append((joint, obj_qpos))
            
            return placed

        placed = None
        for i in range(100):  # 尝试100次
            placed = _find_obj_place_no_contact(find_target_obj=True)
            other_placed = _find_obj_place_no_contact(find_target_obj=False)
            if placed is not None and other_placed is not None:
                placed.extend(other_placed)
                break

        # 容错处理
        if placed is None:
            print(f"Warning: Failed to place objects! Falling back to default positions.")
            placed = [(joint, obj_qpos) for joint, obj_qpos in zip(obj_joints, [np.array([0, 0, 0, 1, 0, 0, 0])] * len(obj_joints))]

        # 一次性写回
        qpos_dict = {jn: q for jn, q in placed}
        env.set_joint_qpos(qpos_dict)
        env.mj_forward()

        self.randomized_object_positions = env.query_joint_qpos(obj_joints)
        self.randomized_goal_positions   = env.query_joint_qpos(goal_joints)


    def generate_actors(self):
        '''
        将所有的 actors 添加到场景中, 初始化到infinity位置
        '''
        for i in range(len(self.actors)):
            self.scene.add_actor(actor_name=self.actors[i], asset_path=self.actors_spawnable[i],
                                 position=np.array(self.infinity), rotation = rotations.euler2quat([0, 0, 0]))
    
    def generate_lights(self):
        """
        只在 random_light=True 时运行，
        并且固定选 self.num_lights 盏，生成朝下的灯。
        """
        idxs = []
        if not self.random_light:
            return idxs

        total = len(self.lights)
        # 限制不超过总数
        n = min(self.num_lights, total)
        idxs = random.sample(range(total), n)

        for i in idxs:
            name      = self.lights[i]
            spawnable = self.lights_spawnable[i]
            print(f"[Debug-Light] add_light → {name} at spawnable {spawnable}")
            self.add_light(name, spawnable)
        return idxs

    def set_actors(self, actors, actors_spawnable, position, rotation):
        for i in range(actors):
            self.add_actor_with_pose(actors[i], actors_spawnable[i], position[i], rotation[i])

    def add_actor(self, actor_name, asset_path):
        random_pos, random_xquat = self._random_position_and_rotation(self.center, self.bound)

        if self.random_actor_position:
            position = random_pos
        else:
            position = self.infinity    # 初始化到无限远位置

        if self.random_actor_rotation:
            rotation = random_xquat
        else:
            rotation = rotations.euler2quat([0, 0, 0])

        self.scene.add_actor(actor_name, asset_path, position, rotation, scale=1.0)


    def add_light(self, light_name, asset_path):
        # 1) 位置：要么随机，要么固定在 center
        rand_pos, rand_quat = self._random_position_and_rotation(self.light_center, self.light_bound)
        position = rand_pos   if self.random_light_position else np.array(self.light_center)

        # 2) 旋转：要么随机，要么固定“竖直向下”
        if self.random_light_rotation:
            rotation = rand_quat
        else:
            # 绕 X 轴 -90° → 光照沿 −Z 方向（竖直向下）
            rotation = rotations.euler2quat(np.array([-np.pi/2, 0.0, 0.0]))

        self.scene.add_light(light_name, asset_path, position, rotation, scale=1.0)

    def set_light_info(self, light_name, color=None, intensity=None):
        """
        Set the light information for a specific light actor.
        :param light_name: The name of the light actor.
        :param color: The color of the light (optional).
        :param intensity: The intensity of the light (optional).
        """
        print(f"[Debug-Light] set_light_info → {light_name}, color={color}, intensity={intensity}")
        if color is None:
            color = np.array([np.random.uniform(0.0, 1.0),
                              np.random.uniform(0.0, 1.0),
                              np.random.uniform(0.0, 1.0)])
        if intensity is None:
            intensity = np.random.uniform(1000, 2000.0)

        self.scene.set_light_info(light_name, color, intensity)

    def set_actor_material(self, actor_name, base_color=None):
        """
        Set the material information for a specific actor.
        :param actor_name: The name of the actor.
        :param base_color: The base color of the material (optional).
        """
        if base_color is None:
            base_color = np.array([np.random.uniform(0.0, 1.0),
                                   np.random.uniform(0.0, 1.0),
                                   np.random.uniform(0.0, 1.0),
                                   1.0])
        print(f"[Debug-Actor] set_actor_material → {actor_name}, base_color={base_color}")
        self.scene.set_material_info(actor_name, base_color)

    def add_actor_with_pose(self, actor_name, asset_path, position, rotation):
        self.scene.add_actor(actor_name, asset_path, position=position, rotation=rotation, scale=1.0)

    def publish_scene(self):
        self.scene.publish_scene()

    def destory_scene(self):
        self.scene.destory_scene()

    def register_init_env_callback(self, init_env_call):
        self.scene.register_init_env_callback(init_env_call)

    def get_task(self, env: OrcaGymLocalEnv):
        raise NotImplementedError("This method should be overridden by subclasses")

    def is_success(self, env: OrcaGymLocalEnv):
        raise NotImplementedError("This method should be overridden by subclasses")

    def get_language_instruction(self) -> str:
        raise NotImplementedError("This method should be overridden by subclasses")

    def __get_config_setting__(self, attr, config: dict):
        if attr in config:
            return config[attr]
        else:
            return DEFAULT_CONFIG[attr]
    
    def __check_config__(self):
        object_len = self.object_bodys.__len__()
        object_joint_len = self.object_joints.__len__()
        object_site_len = self.object_sites.__len__()

        goal_len = self.goal_bodys.__len__()
        goal_joint_len = self.goal_joints.__len__()
        goal_site_len = self.goal_sites.__len__()

    def _random_position_and_rotation(self, center, bound):
        """
        Generate a random position and rotation within the specified bounds.
        """
        position = np.array([np.random.uniform(center[0] + bound[0][0], center[0] + bound[0][1]),
                             np.random.uniform(center[1] + bound[1][0], center[1] + bound[1][1]),
                             np.random.uniform(center[2] + bound[2][0], center[2] + bound[2][1])])
        rotation = rotations.euler2quat(np.array([np.random.uniform(-np.pi, np.pi),
                                                  np.random.uniform(-np.pi, np.pi),
                                                  np.random.uniform(-np.pi, np.pi)]))
        return position, rotation

    def _restore_objects_(self, env: OrcaGymLocalEnv, objects_data):
        """
        恢复物体到指定位置
        :param positions: 物体位置字典
        """
        qpos_dict = {}
        self.object_bodys, self.object_joints = [], []
        if objects_data.shape == () and objects_data.dtype == "object":
            json_str = objects_data[()]
            json_data = json.loads(json_str)
            for object, object_info in json_data.items():
                joint_name = object_info['joint_name']
                qpos_dict[env.joint(joint_name)] = np.array(object_info['position'], dtype=np.float32)
                self.object_bodys.append(object)
                self.object_joints.append(joint_name)
        else:
            arr = objects_data
            empty_name = env.joint("")
            for entry in arr:
                name = entry['joint_name']
                pos = entry['position']
                quat = entry['orientation']
                qpos_dict[name] = np.concatenate([pos, quat], axis=0)
                origin_joint = str(name, encoding='utf-8').removeprefix(empty_name)
                origin_body = origin_joint.removesuffix("_joint")
                self.object_bodys.append(origin_body)
                self.object_joints.append(origin_joint)
        env.set_joint_qpos(qpos_dict)

        env.mj_forward()

    def _restore_goals_(self, env: OrcaGymLocalEnv, goals_data):
        """
        恢复目标到指定位置
        :param positions: 目标位置字典
        """
        qpos_dict = {}
        self.goal_bodys, self.goal_joints = [], []
        if goals_data.shape == () and goals_data.dtype == "object":
            json_str = goals_data[()]
            json_data = json.loads(json_str)
            for object, object_info in json_data.items():
                joint_name = object_info['joint_name']
                qpos_dict[env.joint(joint_name)] = np.array(object_info['position'], dtype=np.float32)
                self.goal_bodys.append(object)
                self.goal_joints.append(joint_name)
        env.set_joint_qpos(qpos_dict)
        env.mj_forward()

    def get_objects_info(self, env: OrcaGymLocalEnv) -> str:
        info = {}
        jpos_dict = self.get_object_joints_xpos(env)
        xpos, xmat, xquat = self.get_object_xpos_xmat_xquat(env)
        for i in range(len(self.object_bodys)):
            info[self.object_bodys[i]] = {
                "joint_name": self.object_joints[i],
                "position": jpos_dict[env.joint(self.object_joints[i])].tolist(),
                "orientation": xquat[i * 4:(i + 1)*4].tolist(),
                "target_body": self.target_object == self.object_bodys[i]
            }
        print(f"info: {info}")

        json_str = json.dumps(info)
        return json_str

    def get_goals_info(self, env: OrcaGymLocalEnv) -> str:
        info = {}
        jpos_dict = self.get_goal_joints_xpos(env)
        xpos, xmat, xquat = self.get_goal_xpos_xmat_xquat(env)
        for i in range(len(self.goal_bodys)):
            geom_size = env.get_goal_bounding_box(env.body(self.goal_bodys[i]))
            if not geom_size:
                continue
            mn = np.array(geom_size["min"]).flatten()
            mx = np.array(geom_size["max"]).flatten()
            sz = mx - mn
            info[self.goal_bodys[i]] = {
                "joint_name": self.goal_joints[i],
                "position": jpos_dict[env.joint(self.goal_joints[i])].tolist(),
                "orientation": xquat[i * 4:(i + 1)*4].tolist(),
                "min": mn.tolist(),
                "max": mx.tolist(),
                "size": sz.tolist()
            }

        json_str = json.dumps(info)
        return json_str

    def resample_objects(self, env: OrcaGymLocalEnv, sample_range: float = 0.0):
        """
        Resample objects within a specified range.
        :param sample_range: The range within which to resample objects.
        """
        if sample_range <= 0.0:
            return

        env_joints = [env.joint(joint_name) for joint_name in self.object_joints]
        joints_pos = env.query_joint_qpos(env_joints)
        for joint, qpos in joints_pos.items():
            new_pos = qpos[:2] + np.random.uniform(-sample_range, sample_range, size=2)
            new_qpos = np.concatenate([new_pos, qpos[2:]])
            joints_pos[joint] = new_qpos

        env.set_joint_qpos(joints_pos)
        env.mj_forward()

    def _set_target_object_(self, env: OrcaGymLocalEnv, data: dict):
        qpos_dict = {}
        objects_data = data['objects']
        if objects_data.shape == () and objects_data.dtype == "object":
            json_str = objects_data[()]
            json_data = json.loads(json_str)
            for object, object_info in json_data.items():
                if object_info["target_body"]:
                    self.target_object = object
        #老数据兼容
        else:
            lang_instr = data.get("language_instruction", b"")
            if isinstance(lang_instr, (bytes, bytearray)):
                lang_instr = lang_instr.decode("utf-8")
            obj_match = re.search(r'object:\s*([^\s]+)', lang_instr)
            self.target_object = obj_match.group(1) if obj_match else None
